# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
import os
import re
from abc import abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Iterable

import torch
from omegaconf import DictConfig
from torch import Tensor

from nemo.collections.asr.inference.model_wrappers.asr_inference_wrapper import ASRInferenceWrapper
from nemo.collections.asr.inference.pipelines.pipeline_interface import PipelineInterface
from nemo.collections.asr.inference.streaming.buffering.audio_bufferer import BatchedAudioBufferer
from nemo.collections.asr.inference.streaming.buffering.cache_feature_bufferer import BatchedCacheFeatureBufferer
from nemo.collections.asr.inference.streaming.buffering.feature_bufferer import BatchedFeatureBufferer
from nemo.collections.asr.inference.streaming.framing.multi_stream import ContinuousBatchedRequestStreamer
from nemo.collections.asr.inference.streaming.framing.request import FeatureBuffer, Frame, Request
from nemo.collections.asr.inference.streaming.framing.request_options import ASRRequestOptions
from nemo.collections.asr.inference.streaming.state.state import StreamingState
from nemo.collections.asr.inference.streaming.text.text_processing import StreamingTextProcessor
from nemo.collections.asr.inference.utils.bpe_decoder import BPEDecoder
from nemo.collections.asr.inference.utils.context_manager import CacheAwareContextManager
from nemo.collections.asr.inference.utils.enums import RequestType
from nemo.collections.asr.inference.utils.pipeline_utils import (
    check_existance_of_required_attributes,
    get_leading_punctuation_regex_pattern,
    ids_to_text_without_stripping,
)
from nemo.collections.asr.inference.utils.progressbar import ProgressBar
from nemo.collections.asr.inference.utils.text_segment import TextSegment
from nemo.collections.common.tokenizers.tokenizer_spec import TokenizerSpec

if TYPE_CHECKING:
    from nemo.collections.asr.inference.itn.inverse_normalizer import AlignmentPreservingInverseNormalizer
    from nemo.collections.asr.inference.nmt.llm_translator import LLMTranslator


@dataclass
class TranscribeStepOutput:
    """
    Stores the output of a single transcribe step.
    """

    stream_id: int
    # Final transcript is the transcript generated started from the previous EoU to the current EoU
    # It is finalized transcript, optionally punctuated and ITN-normalized. It's not subject to further modifications.
    # Final segments contains metadata for each word/segment in the final transcript.
    final_transcript: str = ""
    final_segments: list[TextSegment] | None = None
    final_translation: str = ""
    # Partial transcript is the transcript generated started from the previous EoU up to the current frame
    # It is not finalized transcript, it may be subject to further modifications.
    # It can also contain transcript from future frames.
    partial_transcript: str = ""
    partial_translation: str = ""
    # Current step transcript/translation is the transcript/translation generated from the current frame
    current_step_transcript: str = ""
    current_step_translation: str = ""

    @classmethod
    def from_state(cls, state: StreamingState, request: Request, sep: str = ' ') -> 'TranscribeStepOutput':
        """
        Create a TranscribeStepOutput from a StreamingState
        Args:
            state (StreamingState): The state to create the output from.
            request (Request): The request to create the output from.
            sep (str): The separator for the text postprocessor.
        Returns:
            TranscribeStepOutput: The output for the step.
        """
        final_transcript = state.final_transcript.strip()
        final_segments = [seg.copy() for seg in state.final_segments]
        if len(final_segments) > 0:
            final_segments[0].text = final_segments[0].text.lstrip(sep)
            final_segments[-1].text = final_segments[-1].text.rstrip(sep)

        if final_transcript:
            separator = ''
            if not request.is_first and state.concat_with_space:
                separator = sep
            final_transcript = separator + final_transcript
            if len(final_segments) > 0:
                final_segments[0].text = separator + final_segments[0].text
        return cls(
            stream_id=request.stream_id,
            final_transcript=final_transcript,
            final_segments=final_segments,
            partial_transcript=state.partial_transcript,
            current_step_transcript=state.current_step_transcript,
        )

    def __str__(self) -> str:
        """
        Return a string representation of the TranscribeStepOutput
        """
        info = {
            "final_transcript": self.final_transcript,
            "final_translation": self.final_translation,
            "partial_transcript": self.partial_transcript,
            "partial_translation": self.partial_translation,
            "current_step_transcript": self.current_step_transcript,
        }
        return json.dumps(info, indent=4, ensure_ascii=False)


class BasePipeline(PipelineInterface):
    """
    Base class for all pipelines.
    """

    def __init__(self):
        """Initialize state pool to store the state for each stream"""
        self._state_pool: dict[int, StreamingState] = {}

    def get_state(self, stream_id: int) -> StreamingState:
        """Retrieve state for a given stream ID."""
        return self._state_pool.get(stream_id, None)

    def get_states(self, stream_ids: Iterable[int]) -> list[StreamingState]:
        """Retrieve states for a list of stream IDs."""
        return [self.get_state(stream_id) for stream_id in stream_ids]

    def delete_state(self, stream_id: int) -> None:
        """Delete the state from the state pool."""
        if stream_id in self._state_pool:
            del self._state_pool[stream_id]

    def delete_states(self, stream_ids: Iterable[int]) -> None:
        """Delete states for a list of stream IDs."""
        for stream_id in stream_ids:
            self.delete_state(stream_id)

    def init_state(self, stream_id: int, options: ASRRequestOptions) -> StreamingState:
        """Initialize the state of the stream"""
        if stream_id not in self._state_pool:
            state = self.create_state(options)
            self._state_pool[stream_id] = state
        return self._state_pool[stream_id]

    def reset_session(self) -> None:
        """Reset the frame buffer and internal state pool"""
        self._state_pool.clear()

    def open_session(self) -> None:
        """Start a new session by resetting the internal state pool"""
        self.reset_session()

    def close_session(self) -> None:
        """Close the session by resetting the internal state pool"""
        self.reset_session()

    @abstractmethod
    def transcribe_step_for_frames(self, frames: list[Frame]) -> None:
        """Transcribe a step for frames"""
        pass

    @abstractmethod
    def transcribe_step_for_feature_buffers(self, fbuffers: list[FeatureBuffer]) -> None:
        """Transcribe a step for feature buffers"""
        pass

    @abstractmethod
    def get_request_generator(self) -> ContinuousBatchedRequestStreamer:
        """Return the request generator."""
        pass

    @abstractmethod
    def get_sep(self) -> str:
        """Return the separator for the text postprocessor."""
        pass

    def translate_step(self, states: list[StreamingState], step_outputs: list[TranscribeStepOutput]) -> None:
        """
        Translate step
        Args:
            states (list[StreamingState]): List of StreamingState objects.
            step_outputs (list[TranscribeStepOutput]): List of TranscribeStepOutput objects.
        """
        src_langs, tgt_langs = [], []
        asr_transcripts, current_prefixes, previous_translations = [], [], []
        final_transcript_mask = []
        states_to_translate = []

        src_contexts, tgt_contexts = [], []
        for state, step_output in zip(states, step_outputs):
            if not state.options.enable_nmt:
                continue

            src_lang = state.options.source_language
            tgt_lang = state.options.target_language
            if not src_lang or not tgt_lang:
                raise ValueError("Source and target languages must be set when NMT is enabled")

            final = step_output.final_transcript
            partial = step_output.partial_transcript
            if not (final.strip() or partial.strip()):
                continue

            transcript = final or partial
            is_final = bool(final)
            prev_translation, prefix = state.previous_translation_info

            states_to_translate.append((state, step_output))
            src_langs.append(src_lang)
            tgt_langs.append(tgt_lang)
            asr_transcripts.append(transcript)
            current_prefixes.append(prefix)
            previous_translations.append(prev_translation)
            final_transcript_mask.append(is_final)

            src_context, tgt_context = state.previous_context
            src_contexts.append(src_context)
            tgt_contexts.append(tgt_context)

        if len(states_to_translate) == 0:
            return

        translations = self.nmt_model.translate(
            asr_transcripts, current_prefixes, src_langs, tgt_langs, src_contexts, tgt_contexts
        )
        new_prefixes = self.nmt_model.get_prefixes(asr_transcripts, translations, previous_translations)

        for (state, step_output), translation, new_prefix, prev_prefix, is_final in zip(
            states_to_translate, translations, new_prefixes, current_prefixes, final_transcript_mask
        ):
            if is_final:
                step_output.final_translation = translation
                step_output.partial_translation = ""
                state.cleanup_translation_info_after_eou()
                state.set_translation_context(step_output.final_transcript, translation)
                new_prefix = translation
            else:
                step_output.partial_translation = translation
                step_output.final_translation = ""
                state.set_translation_info(translation, new_prefix)

            lcp = os.path.commonprefix([prev_prefix, new_prefix])
            step_output.current_step_translation = new_prefix[len(lcp) :]

    def transcribe_step(self, requests: list[Request]) -> list[TranscribeStepOutput]:
        """
        Transcribe step
        Args:
            requests (list[Request]): List of Request objects.
        Returns:
            list[TranscribeStepOutput]: List of TranscribeStepOutput objects.
        """

        # Initialize the state if it is the first request for the stream
        states = []
        for request in requests:
            if request.is_first:
                self.init_state(request.stream_id, request.options)
            states.append(self.get_state(request.stream_id))

        # Perform the transcribe step for the frames or feature buffers
        if isinstance(requests[0], Frame):
            self.transcribe_step_for_frames(frames=requests)
        elif isinstance(requests[0], FeatureBuffer):
            self.transcribe_step_for_feature_buffers(fbuffers=requests)
        else:
            raise ValueError(f"Invalid request type: {type(requests[0])}")

        # Create current step output for each request
        outputs = []
        sep = self.get_sep()
        for request, state in zip(requests, states):
            step_output = TranscribeStepOutput.from_state(state=state, request=request, sep=sep)
            outputs.append(step_output)

        # Perform the translation step
        if self.nmt_enabled:
            self.translate_step(states=states, step_outputs=outputs)

        # Cleanup the states after the response is sent
        # If last request, delete state from the state pool to free memory
        for state, request in zip(states, requests):
            state.cleanup_after_response()
            if request.is_last:
                self.delete_state(request.stream_id)
        return outputs

    def copy_asr_model_attributes(self, asr_model: ASRInferenceWrapper) -> None:
        """
        Copy the attributes from the ASR model
        Args:
            asr_model (ASRInferenceWrapper): ASR model to copy the attributes from.
        """
        self.asr_model = asr_model
        self.tokenizer = asr_model.tokenizer
        self.device = asr_model.device
        self.supports_punctuation = asr_model.supports_punctuation()
        self.asr_supported_puncts = asr_model.supported_punctuation()
        self.leading_regex_pattern = get_leading_punctuation_regex_pattern(self.asr_supported_puncts)
        self.blank_id = asr_model.get_blank_id()
        self.vocabulary = asr_model.get_vocabulary()
        self.sep = asr_model.word_separator
        self.underscore_id = asr_model.underscore_id
        self.punctuation_ids = asr_model.punctuation_ids
        self.language_token_ids = asr_model.language_token_ids
        self.preprocessor, self.preprocessor_config = asr_model.create_preprocessor()
        self.subsampling_factor = asr_model.get_subsampling_factor()
        self.window_stride = asr_model.get_window_stride()
        self.model_stride_in_secs = asr_model.get_model_stride(in_secs=True)
        self.model_stride_in_milliseconds = asr_model.get_model_stride(in_milliseconds=True)

    def update_partial_transcript(
        self, requests: list[Request], tokenizer: TokenizerSpec, leading_regex_pattern: str
    ) -> None:
        """
        Update partial and current step transcripts from the state.
        Args:
            requests (list[Request]): List of Request objects.
            tokenizer (TokenizerSpec): Used to convert tokens into text
            leading_regex_pattern (str): Regex pattern for the punctuation marks.
        """
        word_separator = self.get_sep()
        for request in requests:
            state = self.get_state(request.stream_id)
            # state tokens represent all tokens accumulated since the EOU
            # incomplete segment tokens are the remaining tokens on the right side of the buffer after EOU
            all_tokens = state.tokens + state.incomplete_segment_tokens
            if len(all_tokens) > 0:
                pt_string = ids_to_text_without_stripping(all_tokens, tokenizer, word_separator)
                if leading_regex_pattern:
                    pt_string = re.sub(leading_regex_pattern, r'\1', pt_string)
                state.partial_transcript = pt_string
            else:
                state.partial_transcript = ""

            current_step_tokens = state.current_step_tokens
            if len(current_step_tokens) > 0:
                step_transcript = ids_to_text_without_stripping(current_step_tokens, tokenizer, word_separator)
                state.current_step_transcript = step_transcript
            else:
                state.current_step_transcript = ""

    def init_bpe_decoder(self) -> None:
        """Initialize the BPE decoder"""
        check_existance_of_required_attributes(
            self,
            [
                'vocabulary',
                'tokenizer',
                'confidence_aggregator',
                'asr_supported_puncts',
                'word_boundary_tolerance',
                'model_stride_in_secs',
            ],
        )

        self.bpe_decoder = BPEDecoder(
            vocabulary=self.vocabulary,
            tokenizer=self.tokenizer,
            confidence_aggregator=self.confidence_aggregator,
            asr_supported_puncts=self.asr_supported_puncts,
            word_boundary_tolerance=self.word_boundary_tolerance,
            token_duration_in_secs=self.model_stride_in_secs,
        )

    def init_text_processor(
        self,
        cfg: DictConfig,
        itn_model: AlignmentPreservingInverseNormalizer | None,
    ) -> None:
        """
        Initialize the text processor.
        Args:
            cfg: (DictConfig) Configuration parameters.
            itn_model: (AlignmentPreservingInverseNormalizer | None) Inverse Text Normalization model.
        """
        check_existance_of_required_attributes(
            self,
            [
                'asr_supported_puncts',
                'supports_punctuation',
                'confidence_aggregator',
                'sep',
            ],
        )

        self.text_processor = StreamingTextProcessor(
            itn_cfg=cfg.itn,
            itn_model=itn_model,
            asr_supported_puncts=self.asr_supported_puncts,
            asr_supports_punctuation=self.supports_punctuation,
            confidence_aggregator=self.confidence_aggregator,
            sep=self.sep,
            enable_pnc=cfg.enable_pnc,
            enable_itn=cfg.enable_itn,
        )

    def init_nmt_model(self, nmt_model: LLMTranslator | None) -> None:
        """
        Initialize the Translation model.
        Args:
            nmt_model: (LLMTranslator | None) LLM based translation model.
        """
        self.nmt_model = nmt_model
        self.nmt_enabled = nmt_model is not None

    def init_bufferer_for_buffered_streaming(self) -> None:
        """Initialize the bufferer."""
        check_existance_of_required_attributes(
            self,
            [
                'request_type',
                'sample_rate',
                'buffer_size_in_secs',
                'preprocessor_config',
                'device',
            ],
        )

        if self.request_type is RequestType.FEATURE_BUFFER:
            # Feature buffering: It will be used when the input is feature buffers
            self.bufferer = BatchedFeatureBufferer(
                sample_rate=self.sample_rate,
                buffer_size_in_secs=self.buffer_size_in_secs,
                preprocessor_cfg=self.preprocessor_config,
                device=self.device,
            )
        elif self.request_type is RequestType.FRAME:
            # Audio buffering: It will be used when the input is audio frames
            self.bufferer = BatchedAudioBufferer(
                sample_rate=self.sample_rate, buffer_size_in_secs=self.buffer_size_in_secs
            )
        else:
            raise ValueError(f"Unknown request type: {self.request_type}")

    def init_bufferer_for_cache_aware_streaming(self) -> None:
        """Initialize the bufferer for cache-aware streaming."""
        check_existance_of_required_attributes(
            self,
            [
                'num_slots',
                'use_feat_cache',
                'chunk_size_in_secs',
                'buffer_size_in_secs',
                'sample_rate',
                'preprocessor_config',
                'device',
            ],
        )

        if self.use_feat_cache:
            # Only calculate mel-spec features for last chunk
            chunk_size_for_feature_buffer = self.chunk_size_in_secs
        else:
            # Calculate mel-spec features for the whole buffer
            chunk_size_for_feature_buffer = self.buffer_size_in_secs

        self.bufferer = BatchedCacheFeatureBufferer(
            num_slots=self.num_slots,
            sample_rate=self.sample_rate,
            buffer_size_in_secs=self.buffer_size_in_secs,
            chunk_size_in_secs=chunk_size_for_feature_buffer,
            preprocessor_cfg=self.preprocessor_config,
            device=self.device,
        )

    def init_context_manager(self) -> None:
        """Initialize the context manager."""
        check_existance_of_required_attributes(self, ['asr_model', 'num_slots', 'use_cache'])
        self.context_manager = CacheAwareContextManager(
            cache_aware_model=self.asr_model, num_slots=self.num_slots, use_cache=self.use_cache
        )

    def init_prompt_support(self) -> None:
        """Initialize prompt support for multilingual models."""
        self.prompt_enabled = hasattr(self.asr_model.asr_model, 'concat') and self.asr_model.asr_model.concat

        if self.prompt_enabled:
            self._prompt_config = self._load_prompt_config()

    def _load_prompt_config(self) -> dict:
        """
        Load prompt configuration from model.
        Returns:
            (dict) Prompt configuration containing num_prompts, prompt_dict, and compute_dtype.
        """
        cfg = self.asr_model.asr_model.cfg
        if cfg and hasattr(cfg, 'model_defaults'):
            model_defaults = cfg.model_defaults
            num_prompts = model_defaults.get('num_prompts', None)
            prompt_dict = model_defaults.get('prompt_dictionary', None)

            # Validate and convert types once
            num_prompts_int = int(num_prompts) if num_prompts is not None else 0

            is_dict_like = isinstance(prompt_dict, dict) or (
                hasattr(prompt_dict, 'get') and hasattr(prompt_dict, '__contains__')
            )

            if num_prompts_int > 0 and is_dict_like:
                return {
                    'num_prompts': num_prompts_int,
                    'prompt_dict': prompt_dict,
                    'compute_dtype': getattr(self.asr_model.asr_model, 'dtype', torch.float32),
                }

        return {}

    def _resolve_prompt_index(self, language_code: str) -> int:
        """
        Resolve language_code to a strict prompt index; raise if invalid.
        Args:
            language_code: (str) Language code to resolve (e.g., "en-US", "es-ES").
        Returns:
            (int) Prompt index corresponding to the language code.
        Raises:
            RuntimeError: If prompt configuration is missing.
            ValueError: If language_code is not found in prompt dictionary.
        """
        if not hasattr(self, '_prompt_config') or not self._prompt_config:
            raise RuntimeError("Prompt configuration is missing for a prompt-enabled model.")
        prompt_dict = self._prompt_config['prompt_dict']
        lang_index = prompt_dict.get(language_code, None)
        if lang_index is None:
            raise ValueError(
                f"Language code '{language_code}' not found in prompt dictionary. "
                f"Available languages: {list(prompt_dict.keys())}"
            )
        return lang_index

    def _create_one_hot_prompts(self, indices: Tensor) -> Tensor:
        """
        Create one-hot prompt vectors from indices.
        Args:
            indices: (Tensor) Prompt indices of shape [B].
        Returns:
            (Tensor) One-hot prompt vectors of shape [B, num_prompts].
        """
        num_prompts = self._prompt_config['num_prompts']
        return torch.nn.functional.one_hot(indices, num_classes=num_prompts).to(self._prompt_config['compute_dtype'])

    def _build_prompt_vectors(self, states: list) -> Tensor:
        """
        Build prompt vectors for a batch of states using one-hot encoding.
        Args:
            states: (list) List of streaming states.
        Returns:
            (Tensor) Prompt vectors of shape [B, num_prompts].
        Raises:
            ValueError: If any prompt index is out of range.
        """
        indices = torch.tensor([getattr(s, 'prompt_idx', 0) for s in states], device=self.device, dtype=torch.long)
        num_prompts = self._prompt_config['num_prompts']
        if torch.any((indices < 0) | (indices >= num_prompts)):
            raise ValueError("Found out-of-range prompt index in batch.")
        return self._create_one_hot_prompts(indices)

    def run(
        self,
        audio_filepaths: list[str],
        options: list[ASRRequestOptions] | None = None,
        progress_bar: ProgressBar | None = None,
    ) -> dict:
        """
        Orchestrates reading from audio_filepaths in a streaming manner,
        transcribes them, and packs the results into a PipelineOutput.
        Args:
            audio_filepaths (list[str]): List of audio filepaths to transcribe.
            options (list[ASRRequestOptions] | None): List of RequestOptions for each stream.
            progress_bar (ProgressBar | None): Progress bar to show the progress. Default is None.
        Returns:
            dict: A dictionary containing transcriptions and segments for each stream.
        """
        if progress_bar is not None and not isinstance(progress_bar, ProgressBar):
            raise ValueError("progress_bar must be an instance of ProgressBar.")

        if options is None:
            # Use default options if not provided
            options = [ASRRequestOptions() for _ in audio_filepaths]

        if len(options) != len(audio_filepaths):
            raise ValueError("options must be the same length as audio_filepaths")

        request_generator = self.get_request_generator()
        request_generator.set_audio_filepaths(audio_filepaths, options)
        request_generator.set_progress_bar(progress_bar)

        pipeline_output = {}
        sep = self.get_sep()
        self.open_session()
        for requests in request_generator:
            step_outputs = self.transcribe_step(requests)
            for step_output in step_outputs:
                stream_id = step_output.stream_id
                if stream_id not in pipeline_output:
                    pipeline_output[stream_id] = {
                        "text": "",
                        "translation": "",
                        "segments": [],
                        "audio_filepath": request_generator.get_audio_filepath(stream_id),
                        "translation_segments": [],
                    }

                accumulated_text = pipeline_output[stream_id]["text"]
                accumulated_translation = pipeline_output[stream_id]["translation"]
                final_transcript = step_output.final_transcript
                final_translation = step_output.final_translation
                final_segments = step_output.final_segments
                if not accumulated_text:
                    final_transcript = final_transcript.lstrip(sep)
                    if len(final_segments) > 0:
                        first_segment = final_segments[0]
                        first_segment.text = first_segment.text.lstrip(sep)

                if not accumulated_translation:
                    final_translation = final_translation.lstrip(sep)

                accumulated_text += final_transcript
                accumulated_translation += final_translation
                pipeline_output[stream_id]["text"] = accumulated_text
                pipeline_output[stream_id]["translation"] = accumulated_translation
                pipeline_output[stream_id]["segments"].extend(final_segments)

                if self.nmt_enabled:
                    step_translation = step_output.current_step_translation
                    delay = request_generator.get_elapsed_duration(stream_id)
                    pipeline_output[stream_id]["translation_segments"].append((step_translation, delay))

        self.close_session()
        return pipeline_output



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import List, Tuple

from nemo.collections.llm.tools.auto_configurator.core import utils


GPT_BASED_MODELS = [
    "gpt3",
    "bert",
    "llama",
    "qwen",
    "mixtral",
    "mistral",
    "gemma",
    "nemotron",
    "starcoder",
]


def generate_grid_search_configs(
    base_cfg: dict,
    train_cfg: dict,
) -> Tuple[dict, dict]:
    """Generates the grid of all possible configurations for the given model,
        and stores each different configuration in a yaml file.

    Args:
        base_cfg (dict): base configuration of the model to be trained.
        train_cfg (dict): train configuration of the model to be trained.

    Returns:
        dict: base config.
        dict: generated configs.
    """

    model_name = train_cfg.model_type
    model_size_in_b = train_cfg.model_size_in_b
    path_to_logs = train_cfg.path_to_logs

    # 2 * num_layers is needed because of encoder/decoder architecture.
    multiplier = 1 if model_name in GPT_BASED_MODELS else 2

    seq_length = base_cfg.model.config.seq_length
    num_layers = base_cfg.model.config.num_layers

    act_method = None

    params = _calculate_tp_pp_mbs_grid(
        model_size_in_b=model_size_in_b,
        num_layers=num_layers,
        model_name=model_name,
        seq_length=seq_length,
        train_cfg=train_cfg,
    )

    max_steps = train_cfg.max_steps_per_run
    num_nodes = train_cfg.num_nodes
    valid_tp_pp_list = []
    for tp in params.tp:
        for pp in params.pp:
            for cp in params.cp:
                for ep in params.ep:
                    for mbs in params.mbs:
                        num_gpus = base_cfg.trainer.num_nodes * base_cfg.trainer.devices
                        base_cfg.data.global_batch_size = params.gbs
                        att_heads = base_cfg.model.config.num_attention_heads
                        num_layers = base_cfg.model.config.num_layers
                        model_parallelism = (tp * pp * cp * ep) if (cp and ep) else (tp * pp)
                        mod_gbs = params.gbs % (mbs * num_gpus / model_parallelism)
                        mod_att_heads = att_heads % tp
                        mod_layers = (multiplier * num_layers) % pp
                        mod_cp = cp if cp else 1
                        mod_ep = ep if ep else 1
                        if (
                            mod_gbs == 0
                            and mod_att_heads == 0
                            and mod_layers == 0
                            and (tp, pp, cp, ep) not in valid_tp_pp_list
                            and (mod_cp // mod_ep == mod_cp or mod_ep // mod_cp == mod_ep)
                            and params.min_model_parallel <= model_parallelism <= params.max_model_parallel
                        ):
                            valid_tp_pp_list.append((tp, pp, cp, ep))

    # Generate grid search configs.
    configs = {}
    for tp, pp, cp, ep in valid_tp_pp_list:
        (
            virtual_pipelines,
            act_ckpt_layers,
            num_micro_batches_partial_act_ckpt,
            act_ckpt_layers_per_pipeline,
        ) = _set_activations_checkpoint_params(
            tp,
            pp,
            cp,
            ep,
            num_layers,
            act_method,
            multiplier,
            model_size_in_b,
            model_name,
        )
        for mbs in params.mbs:
            kwargs = {
                "base_cfg": base_cfg,
                "act": None,
                "num_mbs_act": None,
                "act_per_pipe": None,
                "tp": tp,
                "pp": pp,
                "cp": cp,
                "ep": ep,
                "virtual_pipelines": virtual_pipelines,
                "mbs": mbs,
                "max_steps": max_steps,
                "num_nodes": num_nodes,
                "path_to_logs": path_to_logs,
                "model_name": model_name,
                "model_size": model_size_in_b,
            }

            act_layers = None
            if act_ckpt_layers[0] is not None:
                if act_layers is not None and act_layers != "auto":
                    act_ckpt_layers = act_layers
                for act in act_ckpt_layers:
                    for num_mbs_act in num_micro_batches_partial_act_ckpt:
                        for act_per_pipe in act_ckpt_layers_per_pipeline:
                            kwargs["act"] = act
                            kwargs["num_mbs_act"] = num_mbs_act
                            kwargs["act_per_pipe"] = act_per_pipe
                            new_cfg = utils.modify_cfg(**kwargs)
                            if new_cfg:  # Save candidate cfg.
                                configs[new_cfg["name"]] = new_cfg
            else:
                new_cfg = utils.modify_cfg(**kwargs)
                if new_cfg:  # Save candidate cfg.
                    config_name = new_cfg["name"]
                    configs[config_name] = new_cfg

    print(f"\nAll candidate configurations created correctly. Total number of configs: {len(configs)}.\n")
    return base_cfg, configs


def _set_activations_checkpoint_params(
    tp, pp, cp, ep, num_layers, act_method, multiplier, model_size_in_b, model_name
):
    act_multiple = 4 // pp
    if act_method == "block":
        if 1.0 <= model_size_in_b < 11.3:
            act_multiple = 8 // pp
        elif 11.3 <= model_size_in_b < 26.0:
            act_multiple = 16 // pp
        elif 26.0 <= model_size_in_b < 60.0:
            act_multiple = 16 // pp
        elif 60.0 <= model_size_in_b:
            act_multiple = 32 // pp
    act_multiple = max(act_multiple, 1)

    virtual_pipelines = None
    # Num micro batches with partial act ckpt
    min_micro_b = 0  # 0 will not be used, minimum will be set to 1 later in the code.
    max_micro_b = pp
    interval_micro_b = 1
    # Act ckpt layers per pipeline
    min_layers_per_pipe = 0
    max_layers_per_pipe = num_layers
    interval_layers_per_pipe = act_multiple
    if model_name in GPT_BASED_MODELS and pp > 2:  # Interleaved pipeline scheduling.
        virtual_pipelines = num_layers // pp  # TODO: verify that this is the best value.
        act_multiple = 1
        max_micro_b = pp * (virtual_pipelines - 1) + (pp - 1) * 2 + 1
        interval_micro_b = virtual_pipelines * 8
        max_layers_per_pipe = multiplier * num_layers // pp // virtual_pipelines + 1

    (
        act_ckpt_layers,
        num_micro_batches_partial_act_ckpt,
        act_ckpt_layers_per_pipeline,
    ) = ([None], [None], [None])
    if act_method == "block":
        # Act ckpt num layers
        if virtual_pipelines is None:
            act_ckpt_layers = range(0, multiplier * num_layers // pp + 1, act_multiple)
        else:
            act_ckpt_layers = range(0, multiplier * num_layers // pp // virtual_pipelines + 1, act_multiple)

        if pp > 1 and model_name in GPT_BASED_MODELS:
            # Num micro batches with partial act ckpt
            num_micro_batches_partial_act_ckpt = list(range(min_micro_b, max_micro_b + 1, interval_micro_b))
            if num_micro_batches_partial_act_ckpt[0] == 0:
                num_micro_batches_partial_act_ckpt[0] = 1

            # Act ckpt layers per pipeline
            act_ckpt_layers_per_pipeline = range(
                min_layers_per_pipe, max_layers_per_pipe + 1, interval_layers_per_pipe
            )

    return (
        virtual_pipelines,
        act_ckpt_layers,
        num_micro_batches_partial_act_ckpt,
        act_ckpt_layers_per_pipeline,
    )


@dataclass
class GPT3GridSearch:
    """Selects grid search space for TP, PP, CP, EP, MBS parameters for GPT-3 and 80GB GPUs.

    Args:
        model_size_in_b (float): number of parameters in the model.
        valid_pp (List[int]): list of valid Pipeline Parallelism (PP) values for this config.
        seq_length (int): sequence length to use for training.
        gpu_memory_gb (int): size of GPU memory in GB.
    """

    model_size_in_b: int
    valid_pp: List[int]
    seq_length: int
    gpu_memory_gb: int

    tp = [1, 2, 4, 8]
    pp = [1]
    cp = [1]
    ep = [1]
    mbs = [1, 2, 4, 8]

    gbs: int = 1024
    min_model_parallel: int = 1
    max_model_parallel: int = 8

    def init_params(self):
        """Initialize model parallelism parameters"""

        model_size_in_b = self.model_size_in_b
        gpu_memory_gb = self.gpu_memory_gb
        seq_length = self.seq_length

        if gpu_memory_gb == 80:
            if seq_length == 2048:
                if model_size_in_b <= 1.0:
                    self.tp = [1, 2]
                    self.gbs = 256
                elif model_size_in_b <= 4.0:
                    self.tp = [1, 2, 4]
                    self.gbs = 1024
                elif model_size_in_b <= 8.0:
                    self.tp = [1, 2, 4]
                    self.gbs = 2048
                elif model_size_in_b <= 13.0:
                    self.tp = [1, 2, 4, 8]
                    self.gbs = 2048
                elif model_size_in_b <= 23.0:
                    self.tp = [1, 2, 4]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 4]
                    self.mbs = [1, 2, 4]
                    self.min_model_parallel = 4
                    self.max_model_parallel = 8
                    self.gbs = 2048
                elif model_size_in_b <= 45.0:
                    self.tp = [2, 4, 8]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 4]
                    self.mbs = [1, 2, 4]
                    self.min_model_parallel = 8
                    self.max_model_parallel = 32
                    self.gbs = 2048
                elif model_size_in_b <= 95:
                    self.tp = [2, 4, 8]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 8]
                    self.mbs = [1, 2, 4, 8]
                    self.min_model_parallel = 8
                    self.max_model_parallel = 64
                    self.gbs = 2048
                elif model_size_in_b <= 130.0:
                    self.tp = [2, 4, 8]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 16]
                    self.mbs = [1, 2, 4, 8]
                    self.min_model_parallel = 16
                    self.max_model_parallel = 128
                    self.gbs = 2048
                elif model_size_in_b <= 195.0:
                    self.tp = [8]
                    self.pp = [x for x in self.valid_pp if 4 <= x <= 16]
                    self.mbs = [1, 2, 4]
                    self.min_model_parallel = 32
                    self.max_model_parallel = 256
                    self.gbs = 2048
                elif model_size_in_b <= 395.0:
                    self.tp = [8]
                    self.pp = [x for x in self.valid_pp if 8 <= x <= 32]
                    self.mbs = [1, 2, 4]
                    self.min_model_parallel = 64
                    self.max_model_parallel = 512
                    self.gbs = 2048
                elif model_size_in_b <= 790.0:
                    self.tp = [8]
                    self.pp = [x for x in self.valid_pp if 8 <= x <= 100]
                    self.mbs = [1, 2, 4]
                    self.min_model_parallel = 128
                    self.max_model_parallel = 1024
                    self.gbs = 2048
                elif model_size_in_b <= 1100.0:
                    self.tp = [8]
                    self.pp = [x for x in self.valid_pp if 16 <= x <= 130]
                    self.mbs = [1, 2, 4]
                    self.min_model_parallel = 256
                    self.max_model_parallel = 2048
                    self.gbs = 2048
            elif seq_length == 4096:
                if model_size_in_b <= 1.0:
                    self.tp = [1, 2, 4]
                    self.mbs = [1, 2, 4, 8]
                    self.gbs = 128
                elif model_size_in_b <= 4.0:
                    self.tp = [1, 2, 4]
                    self.mbs = [1, 2, 4, 8]
                    self.gbs = 512
                elif model_size_in_b <= 8.0:
                    self.tp = [1, 2, 4]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                    self.mbs = [1, 2, 4]
                    self.gbs = 1024
                elif model_size_in_b <= 13.0:
                    self.tp = [2, 4]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                    self.mbs = [1, 2, 4]
                    self.gbs = 1024
                elif model_size_in_b <= 23.0:
                    self.tp = [4, 8]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                    self.mbs = [1, 2]
                    self.min_model_parallel = 4
                    self.max_model_parallel = 16
                    self.gbs = 1024
                elif model_size_in_b <= 45.0:
                    self.tp = [4, 8]
                    self.pp = [x for x in self.valid_pp if 2 <= x <= 4]
                    self.mbs = [1, 2]
                    self.min_model_parallel = 8
                    self.max_model_parallel = 32
                    self.gbs = 1024
                elif model_size_in_b <= 95:
                    self.tp = [4, 8]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 8]
                    self.mbs = [1, 2]
                    self.min_model_parallel = 8
                    self.max_model_parallel = 64
                    self.gbs = 1024
            elif seq_length == 8192:
                if model_size_in_b <= 1.0:
                    self.tp = [1, 2]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                    self.mbs = [1, 2, 4]
                    self.gbs = 64
                elif model_size_in_b <= 4.0:
                    self.tp = [1, 2, 4]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                    self.mbs = [1, 2, 4]
                    self.gbs = 128
                elif model_size_in_b <= 8.0:
                    self.tp = [2, 4]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                    self.mbs = [1, 2]
                    self.gbs = 256
                elif model_size_in_b <= 13.0:
                    self.tp = [2, 4]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                    self.mbs = [1, 2]
                    self.gbs = 256
                elif model_size_in_b <= 23.0:
                    self.tp = [4, 8]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 4]
                    self.mbs = [1]
                    self.min_model_parallel = 8
                    self.max_model_parallel = 32
                    self.gbs = 256
                elif model_size_in_b <= 45.0:
                    self.tp = [8]
                    self.pp = [x for x in self.valid_pp if 4 <= x <= 8]
                    self.mbs = [1]
                    self.min_model_parallel = 32
                    self.max_model_parallel = 64
                    self.gbs = 256
            elif seq_length == 16384:
                if model_size_in_b <= 1.0:
                    self.tp = [2, 4]
                    self.mbs = [1, 2]
                    self.gbs = 32
                elif model_size_in_b <= 4.0:
                    self.tp = [2, 4]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                    self.mbs = [1]
                    self.gbs = 64
                elif model_size_in_b <= 8.0:
                    self.tp = [2, 4]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                    self.mbs = [1]
                    self.gbs = 128
                elif model_size_in_b <= 13.0:
                    self.tp = [2, 4]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                    self.mbs = [1]
                    self.gbs = 128
                elif model_size_in_b <= 23.0:
                    self.tp = [4, 8]
                    self.pp = [x for x in self.valid_pp if 2 <= x <= 4]
                    self.mbs = [1]
                    self.min_model_parallel = 8
                    self.max_model_parallel = 32
                    self.gbs = 128
            elif seq_length == 32768:
                if model_size_in_b <= 1.0:
                    self.tp = [2, 4]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                    self.mbs = [1]
                    self.gbs = 16
                elif model_size_in_b <= 4.0:
                    self.tp = [2, 4]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                    self.mbs = [1]
                    self.gbs = 32
                elif model_size_in_b <= 8.0:
                    self.tp = [4, 8]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                    self.min_model_parallel = 4
                    self.max_model_parallel = 16
                    self.mbs = [1]
                    self.gbs = 64
                elif model_size_in_b <= 13.0:
                    self.tp = [4, 8]
                    self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                    self.min_model_parallel = 4
                    self.max_model_parallel = 16
                    self.mbs = [1]
                    self.gbs = 64
                elif model_size_in_b <= 23.0:
                    self.tp = [8]
                    self.pp = [x for x in self.valid_pp if 2 <= x <= 4]
                    self.mbs = [1]
                    self.min_model_parallel = 16
                    self.max_model_parallel = 32
                    self.gbs = 64
        elif gpu_memory_gb == 40:
            if model_size_in_b <= 1.0:
                self.tp = [1, 2, 4]
                self.mbs = [1, 2, 4, 8]
                self.gbs = 256
            elif model_size_in_b <= 4.0:
                self.tp = [1, 2, 4, 8]
                self.mbs = [1, 2, 4, 8]
                self.gbs = 1024
            elif model_size_in_b <= 8.0:
                self.tp = [2, 4, 8]
                self.pp = [1, 2]
                self.mbs = [1, 2, 4]
                self.min_model_parallel = 2
                self.gbs = 2048
            elif model_size_in_b <= 13.0:
                self.tp = [4, 8]
                self.pp = [1, 2, 4]
                self.mbs = [1, 2, 4]
                self.min_model_parallel = 4
                self.max_model_parallel = 32
                self.gbs = 2048
            elif model_size_in_b <= 23.0:
                self.tp = [2, 4, 8]
                self.pp = [x for x in self.valid_pp if 1 <= x <= 8]
                self.min_model_parallel = 8
                self.max_model_parallel = 64
                self.gbs = 2048
            elif model_size_in_b <= 45.0:
                self.tp = [4, 8]
                self.pp = [x for x in self.valid_pp if 1 <= x <= 12]
                self.mbs = [1, 2, 4]
                self.min_model_parallel = 16
                self.max_model_parallel = 128
                self.gbs = 2048
            elif model_size_in_b <= 95:
                self.tp = [4, 8]
                self.pp = [x for x in self.valid_pp if 1 <= x <= 16]
                self.mbs = [1, 2, 4]
                self.min_model_parallel = 16
                self.max_model_parallel = 256
                self.gbs = 2048
            elif model_size_in_b <= 130.0:
                self.tp = [4, 8]
                self.pp = [x for x in self.valid_pp if 2 <= x <= 26]
                self.mbs = [1, 2]
                self.min_model_parallel = 32
                self.max_model_parallel = 512
                self.gbs = 2048
            elif model_size_in_b <= 195.0:
                self.tp = [4, 8]
                self.pp = [x for x in self.valid_pp if 2 <= x <= 32]
                self.mbs = [1, 2]
                self.min_model_parallel = 64
                self.max_model_parallel = 1024
                self.gbs = 2048
            elif model_size_in_b <= 395.0:
                self.tp = [4, 8]
                self.pp = [x for x in self.valid_pp if 4 <= x <= 64]
                self.mbs = [1, 2]
                self.min_model_parallel = 128
                self.max_model_parallel = 2048
                self.gbs = 2048
            elif model_size_in_b <= 790.0:
                self.tp = [4, 8]
                self.pp = [x for x in self.valid_pp if 8 <= x <= 128]
                self.mbs = [1, 2]
                self.min_model_parallel = 256
                self.max_model_parallel = 4096
                self.gbs = 2048
            elif model_size_in_b <= 1100.0:
                self.tp = [4, 8]
                self.pp = [x for x in self.valid_pp if 8 <= x <= 192]
                self.mbs = [1, 2]
                self.min_model_parallel = 512
                self.max_model_parallel = 8192
                self.gbs = 2048


@dataclass
class T5GridSearch:
    """Selects grid search space for TP, PP, MBS parameters for T5/mT5 and 80GB GPUs.

    Args:
        model_size_in_b (float): number of parameters in the model.
        valid_pp (List[int]): list of valid Pipeline Parallelism (PP) values for this config.
        seq_length (int): sequence length to use for training.
        gpu_memory_gb (int): size of GPU memory in GB.
    """

    model_size_in_b: int
    seq_length: int
    gpu_memory_gb: int
    valid_pp: List[int]

    tp = [1, 2, 4, 8]
    pp = [1]
    cp = [None]
    ep = [None]
    mbs = [1, 2, 4, 6, 8, 12, 16]

    gbs: int = 1920
    min_model_parallel: int = 1
    max_model_parallel: int = 8

    def init_params(self):
        """Initialize model parallelism parameters"""

        model_size_in_b = self.model_size_in_b
        gpu_memory_gb = self.gpu_memory_gb

        if gpu_memory_gb == 80:
            if model_size_in_b <= 1.0:
                self.tp = [1, 2]
                self.mbs = [16, 32, 64, 128]
                self.gbs = 2048
            elif model_size_in_b <= 4.0:
                self.tp = [1, 2, 4]
                self.mbs = [4, 6, 8, 12, 16, 24, 32, 48]
                self.gbs = 1920
            elif model_size_in_b <= 8.0:
                self.tp = [2, 4, 8]
                self.mbs = [4, 6, 8, 12, 16, 24, 32]
                self.gbs = 1920
            elif model_size_in_b <= 14.5:
                self.tp = [4, 8]
                self.mbs = [2, 4, 6, 8, 12, 16, 24]
                self.gbs = 1920
            elif model_size_in_b <= 25.9:
                self.tp = [4, 8]
                self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                self.mbs = [1, 2, 4, 6, 8]
                self.min_model_parallel = 4
                self.max_model_parallel = 16
                self.gbs = 1920
            elif model_size_in_b <= 43.0:
                self.tp = [4, 8]
                self.pp = [x for x in self.valid_pp if 1 <= x <= 4]
                self.mbs = [1, 2, 4, 6, 8]
                self.min_model_parallel = 8
                self.max_model_parallel = 32
                self.gbs = 1920
            elif model_size_in_b <= 85.5:
                self.tp = [4, 8]
                self.pp = [x for x in self.valid_pp if 2 <= x <= 8]
                self.mbs = [1, 2, 4, 6, 8]
                self.min_model_parallel = 16
                self.max_model_parallel = 64
                self.gbs = 1920
            elif model_size_in_b <= 165.5:
                self.tp = [8]
                self.pp = [x for x in self.valid_pp if 4 <= x <= 16]
                self.mbs = [1, 2, 4, 6]
                self.min_model_parallel = 32
                self.max_model_parallel = 128
                self.gbs = 1920
            elif model_size_in_b <= 250:
                self.tp = [8]
                self.pp = [x for x in self.valid_pp if 4 <= x <= 32]
                self.mbs = [1, 2, 4, 6, 8]
                self.min_model_parallel = 64
                self.max_model_parallel = 256
                self.gbs = 1920
        elif gpu_memory_gb == 40:
            if model_size_in_b <= 1.0:
                self.tp = [1, 2]
                self.mbs = [16, 32, 64, 128]
                self.gbs = 2048
            elif model_size_in_b <= 4.0:
                self.tp = [1, 2, 4]
                self.mbs = [4, 8, 12, 16, 24, 32, 48]
                self.gbs = 1920
            elif model_size_in_b <= 8.0:
                self.tp = [2, 4, 8]
                self.mbs = [4, 6, 8, 12, 16, 24]
                self.gbs = 1920
            elif model_size_in_b <= 14.5:
                self.tp = [4, 8]
                self.pp = [x for x in self.valid_pp if 1 <= x <= 2]
                self.mbs = [2, 4, 6, 8, 12, 16]
                self.min_model_parallel = 4
                self.max_model_parallel = 16
                self.gbs = 1920
            elif model_size_in_b <= 25.9:
                self.tp = [4, 8]
                self.pp = [x for x in self.valid_pp if 1 <= x <= 8]
                self.mbs = [1, 2, 4, 6, 8]
                self.min_model_parallel = 8
                self.max_model_parallel = 32
                self.gbs = 1920
            elif model_size_in_b <= 43.0:
                self.tp = [4, 8]
                self.pp = [x for x in self.valid_pp if 1 <= x <= 8]
                self.mbs = [1, 2, 4, 6, 8]
                self.min_model_parallel = 16
                self.max_model_parallel = 32
                self.gbs = 1920
            elif model_size_in_b <= 85.5:
                self.tp = [8]
                self.pp = [x for x in self.valid_pp if 2 <= x <= 8]
                self.mbs = [1, 2, 4, 6, 8]
                self.min_model_parallel = 32
                self.max_model_parallel = 64
                self.gbs = 1920
            elif model_size_in_b <= 165.5:
                self.tp = [8]
                self.pp = [x for x in self.valid_pp if 4 <= x <= 32]
                self.mbs = [1, 2, 4]
                self.min_model_parallel = 64
                self.max_model_parallel = 128
                self.gbs = 1920
            elif model_size_in_b <= 250:
                self.tp = [8]
                self.pp = [x for x in self.valid_pp if 8 <= x <= 64]
                self.mbs = [1, 2, 4]
                self.min_model_parallel = 128
                self.max_model_parallel = 256
                self.gbs = 1920


@dataclass
class BertGridSearch:
    """Selects grid search space for TP, PP, MBS parameters for BERT and 80GB GPUs.

    Args:
        model_size_in_b (float): number of parameters in the model.
        valid_pp (List[int]): list of valid Pipeline Parallelism (PP) values for this config.
        seq_length (int): sequence length to use for training.
        gpu_memory_gb (int): size of GPU memory in GB.
    """

    model_size_in_b: int
    seq_length: int
    gpu_memory_gb: int
    valid_pp: List[int]

    tp = [1, 2, 4, 8]
    pp = [1]
    cp = [None]
    ep = [None]
    mbs = [1, 2, 4, 6, 8, 12, 16]

    gbs: int = 1920
    min_model_parallel: int = 1
    max_model_parallel: int = 8

    def init_params(self):
        """Initialize model parallelism parameters"""

        model_size_in_b = self.model_size_in_b
        gpu_memory_gb = self.gpu_memory_gb

        if gpu_memory_gb == 80:
            if model_size_in_b <= 1.0:
                self.tp = [1, 2]
                self.gbs = 256
            elif model_size_in_b <= 4.0:
                self.tp = [1, 2, 4]
                self.gbs = 1024
            elif model_size_in_b <= 8.0:
                self.tp = [2, 4, 8]
                self.min_model_parallel = 2
                self.gbs = 2048
            elif model_size_in_b <= 13.0:
                self.tp = [2, 4, 8]
                self.mbs = [1, 2, 3, 4, 6]
                self.min_model_parallel = 2
                self.gbs = 2048
            elif model_size_in_b <= 25.0:
                self.tp = [4, 8]
                self.mbs = [1, 2, 3, 4]
                self.min_model_parallel = 4
                self.gbs = 2048
            elif model_size_in_b <= 46.5:
                self.tp = [4, 8]
                self.pp = [1, 2, 4]
                self.mbs = [1, 2, 3, 4]
                self.min_model_parallel = 4
                self.max_model_parallel = 16
                self.gbs = 2048
            elif model_size_in_b <= 87.5:
                self.tp = [4, 8]
                self.pp = [2, 4, 6, 8]
                self.mbs = [1, 2, 3, 4]
                self.min_model_parallel = 8
                self.max_model_parallel = 32
                self.gbs = 2048
            elif model_size_in_b <= 165.5:
                self.tp = [4, 8]
                self.pp = [4, 6, 8, 16]
                self.mbs = [2, 4, 6, 8]
                self.min_model_parallel = 16
                self.max_model_parallel = 128
                self.gbs = 2048
            elif model_size_in_b <= 250.5:
                self.tp = [8]
                self.pp = [4, 8, 16, 32]
                self.mbs = [1, 2, 3, 4]
                self.min_model_parallel = 32
                self.max_model_parallel = 256
                self.gbs = 2048
            else:
                raise ValueError("No BERT model larger than 250B parameters is supported.")
        elif gpu_memory_gb == 40:
            if model_size_in_b <= 1.0:
                self.tp = [1, 2, 4]
                self.gbs = 256
            elif model_size_in_b <= 4.0:
                self.tp = [1, 2, 4, 8]
                self.gbs = 1024
            elif model_size_in_b <= 8.0:
                self.tp = [2, 4, 8]
                self.mbs = [1, 2, 4]
                self.gbs = 2048
            elif model_size_in_b <= 13.0:
                self.tp = [2, 4, 8]
                self.mbs = [1, 2, 4]
                self.gbs = 2048
            elif model_size_in_b <= 25.0:
                self.tp = [2, 4, 8]
                self.pp = [1, 2]
                self.mbs = [1, 2, 4]
                self.min_model_parallel = 2
                self.max_model_parallel = 16
                self.gbs = 2048
            elif model_size_in_b <= 46.5:
                self.tp = [4, 8]
                self.pp = [1, 2, 4, 8]
                self.mbs = [1, 2, 3]
                self.min_model_parallel = 8
                self.max_model_parallel = 32
                self.gbs = 2048
            elif model_size_in_b <= 87.5:
                self.tp = [4, 8]
                self.pp = [2, 4, 6, 8]
                self.mbs = [1, 2, 3]
                self.min_model_parallel = 16
                self.max_model_parallel = 64
                self.gbs = 2048
            elif model_size_in_b <= 165.5:
                self.tp = [8]
                self.pp = [4, 6, 8, 16]
                self.mbs = [1, 2]
                self.min_model_parallel = 32
                self.max_model_parallel = 256
                self.gbs = 2048
            elif model_size_in_b <= 250.5:
                self.tp = [8]
                self.pp = [8, 16, 32]
                self.mbs = [1, 2]
                self.min_model_parallel = 64
                self.max_model_parallel = 512
                self.gbs = 2048
            else:
                raise ValueError("No BERT model larger than 250B parameters is supported.")


def _calculate_tp_pp_mbs_grid(
    model_size_in_b: float,
    num_layers: int,
    model_name: str,
    seq_length: int,
    train_cfg: dict,
) -> Tuple[int, int, int]:
    """Selects grid search space for TP, PP, MBS parameters for any model,
        and calls the necessary heuristics function accordingly.

    Args:
        model_size_in_b (float): number of parameters in the model.
        num_layers (int): number of layers in the model config.
        model_name (str): name of the model to be used, such as gpt3, t5, mt5...
        seq_length (int): sequence length to use for training.
        train_cfg (dict): config of the model that will be launched.

    Returns:
        dataclass object with model parallelism parameters.

    Raises:
        NotImplementedError: if the model_name is not one of the supported models.
    """

    tp_sizes = train_cfg.tensor_parallel_sizes
    pp_sizes = train_cfg.pipeline_parallel_sizes
    cp_sizes = train_cfg.context_parallel_sizes
    ep_sizes = train_cfg.expert_parallel_sizes
    min_model_parallel_size = train_cfg.min_model_parallel_size
    max_model_parallel_size = train_cfg.max_model_parallel_size
    mbs_sizes = train_cfg.micro_batch_sizes
    gbs_size = train_cfg.global_batch_size
    gpu_memory_gb = train_cfg.gpu_memory_gb
    multiplier = 1 if model_name in GPT_BASED_MODELS else 2
    init_pp = [] if model_name in GPT_BASED_MODELS else [1]
    valid_pp = init_pp + [
        multiplier * x for x in range(1, num_layers + 1) if num_layers % x == 0
    ]  # Only divisors of num_layers are possible.

    kwargs = {
        "model_size_in_b": model_size_in_b,
        "valid_pp": valid_pp,
        "seq_length": seq_length,
        "gpu_memory_gb": gpu_memory_gb,
    }

    if model_name in GPT_BASED_MODELS:
        search_class = GPT3GridSearch
    elif model_name in ["t5", "mt5"]:
        search_class = T5GridSearch
    elif model_name == "bert":
        search_class = BertGridSearch
    else:
        raise NotImplementedError("Model name not implemented.")

    params = search_class(**kwargs)
    params.init_params()

    mode = train_cfg.mode
    # Override the tp, pp, mbs search if indicated in the config params.
    if (tp_sizes is not None and tp_sizes != "auto") or mode == "finetune":
        params.tp = tp_sizes
    if (pp_sizes is not None and pp_sizes != "auto") or mode == "finetune":
        params.pp = pp_sizes
    if (cp_sizes is not None and cp_sizes != "auto") or mode == "finetune":
        params.cp = cp_sizes
    if (ep_sizes is not None and ep_sizes != "auto") or mode == "finetune":
        params.ep = ep_sizes
    if (mbs_sizes is not None and mbs_sizes != "auto") or mode == "finetune":
        params.mbs = mbs_sizes
    if (gbs_size is not None and gbs_size != "auto") or mode == "finetune":
        params.gbs = gbs_size
    if (min_model_parallel_size is not None and min_model_parallel_size != "auto") or mode == "finetune":
        params.min_model_parallel = min_model_parallel_size
    if (max_model_parallel_size is not None and max_model_parallel_size != "auto") or mode == "finetune":
        params.max_model_parallel = max_model_parallel_size

    return params



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import asdict, dataclass, fields

import lightning.pytorch as pl
from lightning.pytorch.callbacks.callback import Callback
from megatron.core import ModelParallelConfig
from megatron.core.distributed import DistributedDataParallelConfig
from megatron.core.optimizer import OptimizerConfig

from nemo.collections.llm.recipes.tp_overlap_configs.userbuffers import TransformerLayerTPOverlapCfg
from nemo.lightning.pytorch.strategies.megatron_strategy import MegatronStrategy, ParallelismConfig
from nemo.utils import logging

try:
    from megatron.core.num_microbatches_calculator import get_micro_batch_size
except (ImportError, ModuleNotFoundError):
    logging.warning("Megatron num_microbatches_calculator not found, using Apex version.")
    from apex.transformer.pipeline_parallel.utils import get_micro_batch_size

try:
    import transformer_engine

    HAVE_TE = True
except (ImportError, ModuleNotFoundError):
    HAVE_TE = False


@dataclass
class _CommOverlapConfig:
    # Tensor parallel communication overlap (experimental)
    tp_comm_overlap: bool = None
    tp_comm_overlap_cfg: dict = None
    tp_comm_bootstrap_backend: str = None
    # Pipeline parallel communication overlap
    overlap_p2p_comm: bool = None
    batch_p2p_comm: bool = None
    # Data parallel communication overlap
    overlap_grad_reduce: bool = None
    overlap_param_gather: bool = None
    overlap_param_gather_with_optimizer_step: bool = None
    align_param_gather: bool = None
    bucket_size: int = None
    # Pipeline bubble overlap
    defer_embedding_wgrad_compute: bool = None
    wgrad_deferral_limit: int = None


class MegatronCommOverlapCallback(Callback):
    """
    A PyTorch Lightning callback to enable communication compute overlap.
    This callback enables the following:
        - tensor parallel communication overlap
        - pipeline parallel communication overlap
        - data parallel communication overlap
        - pipeline bubble overlap

    Args:
        tp_comm_overlap (bool): Enable tensor parallel overlap (experimental)
        tp_comm_overlap_cfg (TransformerLayerTPOverlapCfg): Tensor parallel overlap config
        overlap_p2p_comm (bool): Enable pipeline parallel overlap
        batch_p2p_comm (bool): Batch pipeline parallel send/recv into a single op
        overlap_grad_reduce (bool): Overlap data parallel gradient reduction with compute
        overlap_param_gather (bool): Overlap data parallel parameter gather with compute
        overlap_param_gather_with_optimizer_step (bool): Overlap data parallel parameter gather optimizer step
        align_param_gather (bool): Align data parallel parameter gather across virtual pipeline chunks
        bucket_size (int): The DDP bucket size, controls the data parallel overlap granularity
        defer_embedding_wgrad_compute (bool): Overlap wgrads with the pipeline drain bubble for the last pipeline stage
        wgrad_deferral_limit (int): Limit of how many outstanding wgrads may be overlapped with the pipeline drain
                                    bubble

    Example:
        >>> callback = MegatronCommOverlapCallback(tp_comm_overlap=True)
        >>> trainer = Trainer(callbacks=[callback])
    """

    def __init__(
        self,
        tp_comm_overlap: bool = None,
        tp_comm_overlap_cfg: TransformerLayerTPOverlapCfg = None,
        tp_comm_bootstrap_backend: str = None,
        overlap_p2p_comm: bool = None,
        batch_p2p_comm: bool = None,
        overlap_grad_reduce: bool = None,
        overlap_param_gather: bool = None,
        overlap_param_gather_with_optimizer_step: bool = None,
        align_param_gather: bool = None,
        bucket_size: int = None,
        defer_embedding_wgrad_compute: bool = None,
        wgrad_deferral_limit: int = None,
    ):

        self.user_comm_overlap_cfg = _CommOverlapConfig(
            tp_comm_overlap=tp_comm_overlap,
            tp_comm_overlap_cfg=tp_comm_overlap_cfg,
            tp_comm_bootstrap_backend=tp_comm_bootstrap_backend,
            overlap_p2p_comm=overlap_p2p_comm,
            batch_p2p_comm=batch_p2p_comm,
            overlap_grad_reduce=overlap_grad_reduce,
            overlap_param_gather=overlap_param_gather,
            overlap_param_gather_with_optimizer_step=overlap_param_gather_with_optimizer_step,
            align_param_gather=align_param_gather,
            bucket_size=bucket_size,
            defer_embedding_wgrad_compute=defer_embedding_wgrad_compute,
            wgrad_deferral_limit=wgrad_deferral_limit,
        )

        self.tp_comm_overlap_cfg = None
        self.tp_comm_bootstrap_backend = None
        self.need_tp_overlap_ub_init = False

    def _get_model_comm_overlap_cfgs(
        self,
        parallelism_cfg: ParallelismConfig,
    ) -> _CommOverlapConfig:
        comm_overlap_cfg = _CommOverlapConfig()

        vp_size = parallelism_cfg.virtual_pipeline_model_parallel_size
        if vp_size is None:
            vp_size = 1

        # Optimizations disabled by default, can be overriden by user
        comm_overlap_cfg.tp_comm_overlap = False
        comm_overlap_cfg.tp_comm_overlap_cfg = None
        comm_overlap_cfg.tp_comm_bootstrap_backend = None
        comm_overlap_cfg.defer_embedding_wgrad_compute = False
        comm_overlap_cfg.wgrad_deferral_limit = -1

        # Check if TP overlap can be safely enabled
        if self.user_comm_overlap_cfg.tp_comm_overlap is True:
            if parallelism_cfg.tensor_model_parallel_size < 2:
                logging.warning("Disabling tensor parallel communication overlap due to TP size < 2.")
                self.user_comm_overlap_cfg.tp_comm_overlap = False
            elif not parallelism_cfg.sequence_parallel:
                logging.warning("Disabling tensor parallel communication overlap due to sequence_parallel=False.")
                self.user_comm_overlap_cfg.tp_comm_overlap = False
            elif not HAVE_TE:
                logging.warning("Disabling tensor parallel communication overlap due to TE not detected.")
                self.user_comm_overlap_cfg.tp_comm_overlap = False

        # PP overlap
        if parallelism_cfg.pipeline_model_parallel_size > 1:
            if vp_size > 1:
                comm_overlap_cfg.overlap_p2p_comm = True
                comm_overlap_cfg.batch_p2p_comm = False
            else:
                comm_overlap_cfg.overlap_p2p_comm = False
                comm_overlap_cfg.batch_p2p_comm = True
        else:
            comm_overlap_cfg.overlap_p2p_comm = False
            comm_overlap_cfg.batch_p2p_comm = False

        comm_overlap_cfg = self._override_user_cfgs(comm_overlap_cfg)
        return comm_overlap_cfg

    def _get_optimizer_overlap_cfgs(self, parallelism_cfg: ParallelismConfig) -> _CommOverlapConfig:
        from nemo.utils import AppState

        app_state = AppState()
        data_parallel_size = app_state.data_parallel_size

        vp_size = parallelism_cfg.virtual_pipeline_model_parallel_size
        if vp_size is None:
            vp_size = 1

        comm_overlap_cfg = _CommOverlapConfig()
        comm_overlap_cfg.bucket_size = None
        comm_overlap_cfg.overlap_grad_reduce = False
        comm_overlap_cfg.overlap_param_gather = False
        comm_overlap_cfg.overlap_param_gather_with_optimizer_step = False
        comm_overlap_cfg.align_param_gather = False

        if data_parallel_size > 1:
            comm_overlap_cfg.bucket_size = 128 * 1024 * 1024
            comm_overlap_cfg.overlap_grad_reduce = True
            comm_overlap_cfg.overlap_param_gather = True
            if parallelism_cfg.pipeline_model_parallel_size > 1 and vp_size > 1:
                # Currently disabled due to an issue with checkpointing
                # comm_overlap_cfg.overlap_param_gather_with_optimizer_step = True
                comm_overlap_cfg.align_param_gather = True

        comm_overlap_cfg = self._override_user_cfgs(comm_overlap_cfg)
        return comm_overlap_cfg

    def _apply_cfgs(self, src_cfg, dest_cfg):
        # apply optimizations into dest_cfg
        for field in fields(src_cfg):
            if hasattr(dest_cfg, field.name):
                setattr(dest_cfg, field.name, getattr(src_cfg, field.name))

    def _override_user_cfgs(self, comm_overlap_cfg):
        # override default configs with any user provided configs
        if isinstance(self.user_comm_overlap_cfg, _CommOverlapConfig):
            for field in fields(self.user_comm_overlap_cfg):
                user_value = getattr(self.user_comm_overlap_cfg, field.name)
                if user_value is not None:
                    setattr(comm_overlap_cfg, field.name, user_value)

        return comm_overlap_cfg

    def _check_num_cuda_device_max_connections(self):
        import os

        import torch

        from nemo.utils import AppState

        app_state = AppState()
        tp_size = app_state.tensor_model_parallel_size
        cp_size = app_state.context_parallel_size
        major, _ = torch.cuda.get_device_capability()
        if major > 9:
            if tp_size > 1 or cp_size > 1:
                """
                We need extra connections to avoid serialization of streams,
                so we use the max connections of 32 instead of the default
                device connection of 8.
                """
                if not (os.environ.get('CUDA_DEVICE_MAX_CONNECTIONS', None) == "32"):
                    logging.warning(
                        f"Recommend using CUDA_DEVICE_MAX_CONNECTIONS=32 for best performance \
                        but get {os.environ.get('CUDA_DEVICE_MAX_CONNECTIONS', None)}"
                    )
            else:
                if os.environ.get('CUDA_DEVICE_MAX_CONNECTIONS', None) == "1":
                    logging.warning(
                        f"Recommend unsetting CUDA_DEVICE_MAX_CONNECTIONS for best performance \
                        but get {os.environ.get('CUDA_DEVICE_MAX_CONNECTIONS', None)}"
                    )
        else:
            if tp_size > 1 or cp_size > 1:
                """
                Set the device connection to 1 to enforce the kernel queuing
                order from the host to the execution order on GPU. This is
                needed to schedule a communication kernel before the
                overlapping persistent GEMM kernel. Otherwise, the
                communication kernel will be pushed to the end of the GEMM
                kernel so failing to overlap the kernels.
                """
                if not (os.environ.get('CUDA_DEVICE_MAX_CONNECTIONS', None) == "1"):
                    logging.warning(
                        f"Recommend using CUDA_DEVICE_MAX_CONNECTIONS=1 for best performance \
                        but get {os.environ.get('CUDA_DEVICE_MAX_CONNECTIONS', None)}"
                    )

    def setup(self, trainer: pl.Trainer, pl_module: pl.LightningModule, stage: str) -> None:
        """Apply configs set in comm_overlap_cfg on trainer config."""
        assert isinstance(trainer.strategy, MegatronStrategy), "MegatronCommOverlapCallback requires MegatronStrategy"
        parallelism_cfg = trainer.strategy.parallelism

        if hasattr(trainer.model, "config") and isinstance(trainer.model.config, ModelParallelConfig):
            comm_overlap_cfg = self._get_model_comm_overlap_cfgs(parallelism_cfg)
            self._apply_cfgs(comm_overlap_cfg, trainer.model.config)
            if hasattr(trainer.model, '__io__'):
                self._apply_cfgs(comm_overlap_cfg, trainer.model.__io__.config)

            if trainer.model.config.tp_comm_overlap:
                self.tp_comm_overlap_cfg = comm_overlap_cfg.tp_comm_overlap_cfg
                self.tp_comm_bootstrap_backend = comm_overlap_cfg.tp_comm_bootstrap_backend
                self.need_tp_overlap_ub_init = True

        # Data parallel overlap is only available with the Megatron DDP and Distributed optimizer
        if (
            hasattr(trainer.model.optim, "config")
            and isinstance(trainer.model.optim.config, OptimizerConfig)
            and isinstance(trainer.strategy.ddp_config, DistributedDataParallelConfig)
            and trainer.strategy.ddp_config.use_distributed_optimizer
        ):
            comm_overlap_cfg = self._get_optimizer_overlap_cfgs(parallelism_cfg)
            self._apply_cfgs(comm_overlap_cfg, trainer.model.optim.config)
            self._apply_cfgs(comm_overlap_cfg, trainer.strategy.ddp_config)
            if hasattr(trainer.model, '__io__'):
                self._apply_cfgs(comm_overlap_cfg, trainer.model.__io__.optim.config)

        # setup cuda device max connections
        self._check_num_cuda_device_max_connections()

    def _init_te_userbuffers(self, model_parallel_cfg: ModelParallelConfig):
        from megatron.core import parallel_state

        if self.tp_comm_overlap_cfg is None:
            logging.warning(
                "Tensor parallel overlap: No overlap config provided. Initializing TP comm overlap with the default"
                " config."
            )
        else:
            # ub_cfgs is a dataclass, however TE needs a dict, so convert here
            self.tp_comm_overlap_cfg = asdict(self.tp_comm_overlap_cfg)
            # remove keys with None values from dictionary to match TE's expectations
            self.tp_comm_overlap_cfg = {
                key: value for key, value in self.tp_comm_overlap_cfg.items() if value is not None
            }

        micro_batch_size = get_micro_batch_size()
        hidden_size = model_parallel_cfg.hidden_size
        sequence_length = model_parallel_cfg.seq_length
        fp8 = model_parallel_cfg.fp8 is not None

        input_shape = [
            sequence_length * micro_batch_size // parallel_state.get_context_parallel_world_size(),
            hidden_size,
        ]

        try:
            transformer_engine.pytorch.module.base.initialize_ub(
                shape=input_shape,
                tp_size=parallel_state.get_tensor_model_parallel_world_size(),
                use_fp8=fp8,
                ub_cfgs=self.tp_comm_overlap_cfg,
                bootstrap_backend=self.tp_comm_bootstrap_backend,
            )
        except Exception as error:
            raise Exception(f"Tensor parallel overlap: userbuffer initialization failed with {error}")

        self.need_tp_overlap_ub_init = False

    # _init_te_userbuffers must run once before any stages, however there isnt such a
    # unified callback, so add a hook for every stage
    def on_fit_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Actions before the training stage starts."""
        if self.need_tp_overlap_ub_init:
            self._init_te_userbuffers(trainer.model.config)

    def on_validation_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Actions before the validation stage starts."""
        if self.need_tp_overlap_ub_init:
            self._init_te_userbuffers(trainer.model.config)

    def on_test_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Actions before the test stage starts."""
        if self.need_tp_overlap_ub_init:
            self._init_te_userbuffers(trainer.model.config)

    def on_predict_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Actions before the predict stage starts."""
        if self.need_tp_overlap_ub_init:
            self._init_te_userbuffers(trainer.model.config)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import numpy as np

from nemo.collections.common.tokenizers import TokenizerSpec
from nemo.collections.llm.gpt.data.core import create_sft_dataset
from nemo.utils import logging
from nemo.utils.sequence_packing_utils import create_hist, create_packing_strategy, fill_packing_strategy


def tokenize_dataset(
    path: Path,
    tokenizer: TokenizerSpec,
    max_seq_length: int,
    seed: int,
    dataset_kwargs: Optional[dict],
):
    """
    Tokenizes a dataset from the provided path using the specified tokenizer
    and prepares it for further processing.

    Args:
        path (Path): Path to the dataset file.
        tokenizer (TokenizerSpec): The tokenizer to use for tokenization.
        max_seq_length (int): Maximum sequence length for the tokens.
        seed (int): Random seed for shuffling the dataset (optional).

    Returns:
        np.ndarray: A NumPy array containing the tokenized data.
    """

    if not dataset_kwargs:
        dataset_kwargs = {}

    ts = dataset_kwargs.get("tool_schemas")
    if ts and not isinstance(ts, str):
        dataset_kwargs["tool_schemas"] = json.dumps(ts)

    chat_template = dataset_kwargs.pop("chat_template", None)
    if chat_template:
        # Needs to be called after the trainer has started and populated the tokenizer
        # But it can't be in prepare_data because it is only called in Rank 0
        tokenizer.tokenizer.chat_template = chat_template

    dataset = create_sft_dataset(
        path=path,
        tokenizer=tokenizer,
        seq_length=max_seq_length,
        seed=seed,
        is_test=True,
        **dataset_kwargs,
    )
    return np.array([dataset[i] for i in range(len(dataset))])


def prepare_packed_sequence_data(
    input_path: Path,
    output_path: Path,
    output_metadata_path: Path,
    packed_sequence_size: int,
    tokenizer: TokenizerSpec,
    max_seq_length: int,
    seed: Optional[int] = 0,
    packing_algorithm: str = "first_fit_shuffle",
    dataset_kwargs: dict = None,
):
    """
    Prepares a packed sequence dataset from a given input file and saves it to an output file.

    Args:
        input_path (Path): Path to the input dataset file.
        output_path (Path): Path to save the packed sequence data.
        packed_sequence_size (int): The maximum size for each packed sequence.
        tokenizer (TokenizerSpec): The tokenizer to use for tokenization.
        max_seq_length (int): Maximum sequence length for the tokens.
        seed (Optional[int]): Random seed for shuffling (optional).
        packing_algorithm (str): The algorithm used for packing sequences
                currently supports "first_fit_shuffle" and "first_fit_decreasing".

    Returns:
        None: Saves the packed sequence data to the specified output path.
    """

    logging.info(f"Preparing packed sequence from {input_path}")
    dataset = tokenize_dataset(input_path, tokenizer, max_seq_length, seed, dataset_kwargs)
    sequences, histogram = create_hist(dataset, max_seq_length)

    assignments, packing_metadata = create_packing_strategy(histogram, packed_sequence_size, packing_algorithm)
    output_data = fill_packing_strategy(assignments, sequences, packed_sequence_size, tokenizer.eos_id)

    # save output data
    np.save(output_path, output_data)

    # save packing metadata, packing_metadata is appended to the packing file if it exists
    if output_metadata_path is not None:
        try:
            with open(output_metadata_path, "r") as f:
                packing_metadata_file = json.load(f)
                # 'packing_metadata_file' is expected to be a list of dicts: List[Dict[str, int]]
                # Each dict corresponds to a packed dataset. Typically there will be two dicts,
                # one each for the packed val and train datasets.
                # Each dict records two values: 'max_samples_per_bin', the max
                # number of samples per packed sequence, and 'dataset_max_seqlen', the max
                # sequence length per sample in the packed dataset.
                assert isinstance(packing_metadata_file, list), "invalid packing_metadata_file!"
        except FileNotFoundError:
            packing_metadata_file = []

        packing_metadata_file.append(packing_metadata)
        with open(output_metadata_path, "w") as f:
            json.dump(packing_metadata_file, f)

    logging.info(f"Packed sequence is prepared and saved to {output_path}")


@dataclass
class PackedSequenceSpecs:
    """Defines the packed sequence specifications used for generating a packed dataset."""

    packed_sequence_size: int = -1
    """
    If a positive integer, this arg enables training with sequence packing and specifies the pack size
    If less than or equal to 0, sequence packing is disabled. Defaults to -1.
    Note: This arg is distinct from `seq_length` because `seq_length` specifies the maximum
        length of the original sequence
    (i.e. the length to truncate long sequences in the input data).
    """

    tokenizer_model_name: str = None
    """
    Keep track of tokenizer model name, since each tokenizer produces a different packed sequence dataset file.
    This field is set by llm.finetune api.
    """

    packed_train_data_path: str = None
    """
    If specified, use this file for the packed training dataset instead of the default path.
    """

    packed_val_data_path: str = None
    """
    If specified, use this file for the packed validation dataset instead of the default path.
    """

    packed_metadata_path: str = None
    """
    If specified, use this file for the training and validation packing metadata file instead of the default path.
    """

    pad_cu_seqlens: bool = False
    """
    If True, pad cu_seqlens to a constant size, which is required for use with cudagraphs.
    """

    def __post_init__(self):
        if self.packed_train_data_path is not None:
            self.packed_train_data_path = Path(self.packed_train_data_path)
            assert (
                self.packed_train_data_path.suffix == ".npy"
            ), f"packed training data file must be a .npy file: {self.packed_train_data_path}"
            assert (
                self.packed_train_data_path.exists()
            ), f"packed training data file does not exist: {self.packed_train_data_path}"

        if self.packed_val_data_path is not None:
            self.packed_val_data_path = Path(self.packed_val_data_path)
            assert (
                self.packed_val_data_path.suffix == ".npy"
            ), f"packed validation data file must be a .npy file: {self.packed_val_data_path}"
            assert (
                self.packed_val_data_path.exists()
            ), f"packed validation data file does not exist: {self.packed_val_data_path}"



# Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Dict, Optional, Sequence

import einops
import einops.layers.torch
import torch
import torch.nn.functional as F

from nemo.collections.common.parts.utils import activation_registry, mask_sequence_tensor
from nemo.core.classes import NeuralModule, typecheck
from nemo.core.neural_types import FloatType, LengthsType, NeuralType, SpectrogramType, VoidType
from nemo.utils import logging


class SpectrogramNoiseConditionalScoreNetworkPlusPlus(NeuralModule):
    """This model handles complex-valued inputs by stacking real and imaginary components.
    Stacked tensor is processed using NCSN++ and the output is projected to generate real
    and imaginary components of the output channels.

    Args:
        in_channels: number of input complex-valued channels
        out_channels: number of output complex-valued channels
    """

    def __init__(self, *, in_channels: int = 1, out_channels: int = 1, **kwargs):
        super().__init__()

        # Number of input signals for this estimator
        if in_channels < 1:
            raise ValueError(
                f'Number of input channels needs to be larger or equal to one, current value {in_channels}'
            )

        self.in_channels = in_channels

        # Number of output signals for this estimator
        if out_channels < 1:
            raise ValueError(
                f'Number of output channels needs to be larger or equal to one, current value {out_channels}'
            )

        self.out_channels = out_channels

        # Instantiate noise conditional score network NCSN++
        ncsnpp_params = kwargs.copy()
        ncsnpp_params['in_channels'] = ncsnpp_params['out_channels'] = 2 * self.in_channels  # stack real and imag
        self.ncsnpp = NoiseConditionalScoreNetworkPlusPlus(**ncsnpp_params)

        # Output projection to generate real and imaginary components of the output channels
        self.output_projection = torch.nn.Conv2d(
            in_channels=2 * self.in_channels, out_channels=2 * self.out_channels, kernel_size=1
        )

        logging.debug('Initialized %s with', self.__class__.__name__)
        logging.debug('\tin_channels:  %s', self.in_channels)
        logging.debug('\tout_channels: %s', self.out_channels)

    @property
    def input_types(self) -> Dict[str, NeuralType]:
        """Returns definitions of module output ports."""
        return {
            "input": NeuralType(('B', 'C', 'D', 'T'), SpectrogramType()),
            "input_length": NeuralType(('B',), LengthsType(), optional=True),
            "condition": NeuralType(('B',), FloatType(), optional=True),
        }

    @property
    def output_types(self) -> Dict[str, NeuralType]:
        """Returns definitions of module output ports."""
        return {
            "output": NeuralType(('B', 'C', 'D', 'T'), SpectrogramType()),
            "output_length": NeuralType(('B',), LengthsType(), optional=True),
        }

    @typecheck()
    def forward(self, input, input_length=None, condition=None):
        # Stack real and imaginary components
        B, C_in, D, T = input.shape

        if C_in != self.in_channels:
            raise RuntimeError(f'Unexpected input channel size {C_in}, expected {self.in_channels}')

        # Stack real and imaginary parts
        input_real_imag = torch.stack([input.real, input.imag], dim=2)
        input = einops.rearrange(input_real_imag, 'B C RI F T -> B (C RI) F T')

        # Process using NCSN++
        output, output_length = self.ncsnpp(input=input, input_length=input_length, condition=condition)

        # Output projection
        output = self.output_projection(output)

        # Convert to complex-valued signal
        output = output.reshape(B, 2, self.out_channels, D, T)
        # Move real/imag dimension to the end
        output = output.permute(0, 2, 3, 4, 1)
        output = torch.view_as_complex(output.contiguous())

        return output, output_length


class NoiseConditionalScoreNetworkPlusPlus(NeuralModule):
    """Implementation of Noise Conditional Score Network (NCSN++) architecture.

    References:
        - Song et al., Score-Based Generative Modeling through Stochastic Differential Equations, NeurIPS 2021
        - Brock et al., Large scale GAN training for high fidelity natural image synthesis, ICLR 2018
    """

    def __init__(
        self,
        nonlinearity: str = "swish",
        in_channels: int = 2,  # number of channels in the input image
        out_channels: int = 2,  # number of channels in the output image
        channels: Sequence[int] = (128, 128, 256, 256, 256),  # number of channels at start + at every resolution
        num_res_blocks: int = 2,
        num_resolutions: int = 4,
        init_scale: float = 1e-5,
        conditioned_on_time: bool = False,
        fourier_embedding_scale: float = 16.0,
        dropout_rate: float = 0.0,
        pad_time_to: Optional[int] = None,
        pad_dimension_to: Optional[int] = None,
        **_,
    ):
        # Network topology is a flavor of UNet, example chart for num_resolutions=4
        #
        # 1: Image  → Image/2  → Image/4  → Image/8
        #       ↓        ↓          ↓          ↓
        # 2: Hidden → Hidden/2 → Hidden/4 → Hidden/8
        #       ↓        ↓          ↓          ↓
        # 3: Hidden ← Hidden/2 ← Hidden/4 ← Hidden/8
        #       ↓        ↓          ↓          ↓
        # 4: Image  ← Image/2  ← Image/4  ← Image/8

        # Horizontal arrows in (1) are downsampling
        # Vertical arrows from (1) to (2) are channel upconversions
        #
        # Horizontal arrows in (2) are blocks with downsampling where necessary
        # Horizontal arrows in (3) are blocks with upsampling where necessary
        #
        # Vertical arrows from (1) to (2) are downsampling and channel upconversioins
        # Vertical arrows from (2) to (3) are sums connections (also with / sqrt(2))
        # Vertical arrows from (3) to (4) are channel downconversions
        # Horizontal arrows in (4) are upsampling and addition
        super().__init__()

        # same nonlinearity is used throughout the whole network
        self.activation: torch.nn.Module = activation_registry[nonlinearity]()
        self.init_scale: float = init_scale

        self.downsample = torch.nn.Upsample(scale_factor=0.5, mode="bilinear")
        self.upsample = torch.nn.Upsample(scale_factor=2, mode="bilinear")

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.channels = channels
        self.num_res_blocks = num_res_blocks
        self.num_resolutions = num_resolutions
        self.conditioned_on_time = conditioned_on_time

        # padding setup
        self.pad_time_to = pad_time_to or 2**self.num_resolutions
        self.pad_dimension_to = pad_dimension_to or 2**self.num_resolutions

        if self.conditioned_on_time:
            self.time_embedding = torch.nn.Sequential(
                GaussianFourierProjection(embedding_size=self.channels[0], scale=fourier_embedding_scale),
                torch.nn.Linear(self.channels[0] * 2, self.channels[0] * 4),
                self.activation,
                torch.nn.Linear(self.channels[0] * 4, self.channels[0] * 4),
            )

        self.input_pyramid = torch.nn.ModuleList()
        for ch in self.channels[:-1]:
            self.input_pyramid.append(torch.nn.Conv2d(in_channels=self.in_channels, out_channels=ch, kernel_size=1))

        # each block takes an image and outputs an image
        # possibly changes number of channels
        # output blocks ("reverse" path of the unet) reuse outputs of input blocks ("forward" path)
        # so great care must be taken to in/out channels of each block
        # resolutions are handled in `forward`
        block_params = {
            "activation": self.activation,
            "dropout_rate": dropout_rate,
            "init_scale": self.init_scale,
            "diffusion_step_embedding_dim": channels[0] * 4 if self.conditioned_on_time else None,
        }
        self.input_blocks = torch.nn.ModuleList()
        for in_ch, out_ch in zip(self.channels[:-1], self.channels[1:]):
            for n in range(num_res_blocks):
                block = ResnetBlockBigGANPlusPlus(in_ch=in_ch if n == 0 else out_ch, out_ch=out_ch, **block_params)
                self.input_blocks.append(block)

        self.output_blocks = torch.nn.ModuleList()
        for in_ch, out_ch in zip(reversed(self.channels[1:]), reversed(self.channels[:-1])):
            for n in reversed(range(num_res_blocks)):
                block = ResnetBlockBigGANPlusPlus(in_ch=in_ch, out_ch=out_ch if n == 0 else in_ch, **block_params)
                self.output_blocks.append(block)

        self.projection_blocks = torch.nn.ModuleList()
        for ch in self.channels[:-1]:
            self.projection_blocks.append(torch.nn.Conv2d(ch, out_channels, kernel_size=1))

        assert len(self.input_pyramid) == self.num_resolutions
        assert len(self.input_blocks) == self.num_resolutions * self.num_res_blocks
        assert len(self.output_blocks) == self.num_resolutions * self.num_res_blocks
        assert len(self.projection_blocks) == self.num_resolutions

        self.init_weights_()

        logging.debug('Initialized %s with', self.__class__.__name__)
        logging.debug('\tin_channels:         %s', self.in_channels)
        logging.debug('\tout_channels:        %s', self.out_channels)
        logging.debug('\tchannels:            %s', self.channels)
        logging.debug('\tnum_res_blocks:      %s', self.num_res_blocks)
        logging.debug('\tnum_resolutions:     %s', self.num_resolutions)
        logging.debug('\tconditioned_on_time: %s', self.conditioned_on_time)
        logging.debug('\tpad_time_to:         %s', self.pad_time_to)
        logging.debug('\tpad_dimension_to:    %s', self.pad_dimension_to)

    def init_weights_(self):
        for module in self.modules():
            if isinstance(module, (torch.nn.Linear, torch.nn.Conv2d)):
                torch.nn.init.xavier_uniform_(module.weight)
                if module.bias is not None:
                    torch.nn.init.zeros_(module.bias)

        # torch.nn submodules with scaled init
        for module in self.projection_blocks:
            torch.nn.init.xavier_uniform_(module.weight, gain=self.init_scale)

        # non-torch.nn submodules can have their own init schemes
        for module in self.modules():
            if module is self:
                continue

            if hasattr(module, "init_weights_"):
                module.init_weights_()

    @typecheck(
        input_types={
            "input": NeuralType(('B', 'C', 'D', 'T')),
        },
        output_types={
            "output": NeuralType(('B', 'C', 'D', 'T')),
        },
    )
    def pad_input(self, input: torch.Tensor) -> torch.Tensor:
        """Pad input tensor to match the required dimensions across `T` and `D`."""
        *_, D, T = input.shape
        output = input

        # padding across time
        if T % self.pad_time_to != 0:
            output = F.pad(output, (0, self.pad_time_to - T % self.pad_time_to))

        # padding across dimension
        if D % self.pad_dimension_to != 0:
            output = F.pad(output, (0, 0, 0, self.pad_dimension_to - D % self.pad_dimension_to))

        return output

    @property
    def input_types(self) -> Dict[str, NeuralType]:
        """Returns definitions of module output ports."""
        return {
            "input": NeuralType(('B', 'C', 'D', 'T'), VoidType()),
            "input_length": NeuralType(('B',), LengthsType(), optional=True),
            "condition": NeuralType(('B',), FloatType(), optional=True),
        }

    @property
    def output_types(self) -> Dict[str, NeuralType]:
        """Returns definitions of module output ports."""
        return {
            "output": NeuralType(('B', 'C', 'D', 'T'), VoidType()),
            "output_length": NeuralType(('B',), LengthsType(), optional=True),
        }

    @typecheck()
    def forward(
        self, *, input: torch.Tensor, input_length: Optional[torch.Tensor], condition: Optional[torch.Tensor] = None
    ):
        """Forward pass of the model.

        Args:
            input: input tensor, shjae (B, C, D, T)
            input_length: length of the valid time steps for each example in the batch, shape (B,)
            condition: scalar condition (time) for the model, will be embedded using `self.time_embedding`
        """
        assert input.shape[1] == self.in_channels

        # apply padding at the input
        *_, D, T = input.shape
        input = self.pad_input(input=input)

        if input_length is None:
            # assume all time frames are valid
            input_length = torch.LongTensor([input.shape[-1]] * input.shape[0]).to(input.device)

        lengths = input_length

        if condition is not None:
            if len(condition.shape) != 1:
                raise ValueError(
                    f"Expected conditon to be a 1-dim tensor, got a {len(condition.shape)}-dim tensor of shape {tuple(condition.shape)}"
                )
            if condition.shape[0] != input.shape[0]:
                raise ValueError(
                    f"Condition {tuple(condition.shape)} and input {tuple(input.shape)} should match along the batch dimension"
                )

            condition = self.time_embedding(torch.log(condition))

        # downsample and project input image to add later in the downsampling path
        pyramid = [input]
        for resolution_num in range(self.num_resolutions - 1):
            pyramid.append(self.downsample(pyramid[-1]))
        pyramid = [block(image) for image, block in zip(pyramid, self.input_pyramid)]

        # downsampling path
        history = []
        hidden = torch.zeros_like(pyramid[0])
        input_blocks = iter(self.input_blocks)
        for resolution_num, image in enumerate(pyramid):
            hidden = (hidden + image) / math.sqrt(2.0)
            hidden = mask_sequence_tensor(hidden, lengths)

            for _ in range(self.num_res_blocks):
                hidden = next(input_blocks)(hidden, condition)
                hidden = mask_sequence_tensor(hidden, lengths)
                history.append(hidden)

            final_resolution = resolution_num == self.num_resolutions - 1
            if not final_resolution:
                hidden = self.downsample(hidden)
                lengths = (lengths / 2).ceil().long()

        # upsampling path
        to_project = []
        for residual, block in zip(reversed(history), self.output_blocks):
            if hidden.shape != residual.shape:
                to_project.append(hidden)
                hidden = self.upsample(hidden)
                lengths = (lengths * 2).long()

            hidden = (hidden + residual) / math.sqrt(2.0)
            hidden = block(hidden, condition)
            hidden = mask_sequence_tensor(hidden, lengths)

        to_project.append(hidden)

        # projecting to images
        images = []
        for tensor, projection in zip(to_project, reversed(self.projection_blocks)):
            image = projection(tensor)
            images.append(F.interpolate(image, size=input.shape[-2:]))  # TODO write this loop using self.upsample

        result = sum(images)

        assert result.shape[-2:] == input.shape[-2:]

        # remove padding
        result = result[:, :, :D, :T]
        return result, input_length


class GaussianFourierProjection(NeuralModule):
    """Gaussian Fourier embeddings for input scalars.

    The input scalars are typically time or noise levels.
    """

    def __init__(self, embedding_size: int = 256, scale: float = 1.0):
        super().__init__()
        self.W = torch.nn.Parameter(torch.randn(embedding_size) * scale, requires_grad=False)

    @property
    def input_types(self) -> Dict[str, NeuralType]:
        """Returns definitions of module output ports."""
        return {
            "input": NeuralType(('B',), FloatType()),
        }

    @property
    def output_types(self) -> Dict[str, NeuralType]:
        """Returns definitions of module output ports."""
        return {
            "output": NeuralType(('B', 'D'), VoidType()),
        }

    def forward(self, input):
        x_proj = input[:, None] * self.W[None, :] * 2 * math.pi
        return torch.cat([torch.sin(x_proj), torch.cos(x_proj)], dim=-1)


class ResnetBlockBigGANPlusPlus(torch.nn.Module):
    """Implementation of a ResNet block for the BigGAN model.

    References:
        - Song et al., Score-Based Generative Modeling through Stochastic Differential Equations, NeurIPS 2021
        - Brock et al., Large scale GAN training for high fidelity natural image synthesis, ICLR 2018
    """

    def __init__(
        self,
        activation: torch.nn.Module,
        in_ch: int,
        out_ch: int,
        diffusion_step_embedding_dim: Optional[int] = None,
        init_scale: float = 1e-5,
        dropout_rate: float = 0.1,
        in_num_groups: Optional[int] = None,
        out_num_groups: Optional[int] = None,
        eps: float = 1e-6,
    ):
        """
        Args:
            activation (torch.nn.Module): activation layer (ReLU, SiLU, etc)
            in_ch (int): number of channels in the input image
            out_ch (int, optional): number of channels in the output image
            diffusion_step_embedding_dim (int, optional): dimension of diffusion timestep embedding. Defaults to None (no embedding).
            dropout_rate (float, optional): dropout rate. Defaults to 0.1.
            init_scale (float, optional): scaling for weight initialization. Defaults to 0.0.
            in_num_groups (int, optional): num_groups in the first GroupNorm. Defaults to min(in_ch // 4, 32)
            out_num_groups (int, optional): num_groups in the second GroupNorm. Defaults to min(out_ch // 4, 32)
            eps (float, optional): eps parameter of GroupNorms. Defaults to 1e-6.
        """
        super().__init__()
        in_num_groups = in_num_groups or min(in_ch // 4, 32)
        out_num_groups = out_num_groups or min(out_ch // 4, 32)

        self.init_scale = init_scale

        self.input_block = torch.nn.Sequential(
            torch.nn.GroupNorm(num_groups=in_num_groups, num_channels=in_ch, eps=eps),
            activation,
        )

        self.middle_conv = torch.nn.Conv2d(in_channels=in_ch, out_channels=out_ch, kernel_size=3, padding=1)
        if diffusion_step_embedding_dim is not None:
            self.diffusion_step_projection = torch.nn.Sequential(
                activation,
                torch.nn.Linear(diffusion_step_embedding_dim, out_ch),
                einops.layers.torch.Rearrange("batch dim -> batch dim 1 1"),
            )

        self.output_block = torch.nn.Sequential(
            torch.nn.GroupNorm(num_groups=out_num_groups, num_channels=out_ch, eps=eps),
            activation,
            torch.nn.Dropout(dropout_rate),
            torch.nn.Conv2d(in_channels=out_ch, out_channels=out_ch, kernel_size=3, padding=1),
        )

        if in_ch != out_ch:
            self.residual_projection = torch.nn.Conv2d(in_channels=in_ch, out_channels=out_ch, kernel_size=1)

        self.act = activation
        self.in_ch = in_ch
        self.out_ch = out_ch

        self.init_weights_()

    def init_weights_(self):
        """Weight initialization"""
        for module in self.modules():
            if isinstance(module, (torch.nn.Conv2d, torch.nn.Linear)):
                torch.nn.init.xavier_uniform_(module.weight)
                if module.bias is not None:
                    torch.nn.init.zeros_(module.bias)

        # a single Conv2d is initialized with gain
        torch.nn.init.xavier_uniform_(self.output_block[-1].weight, gain=self.init_scale)

    def forward(self, x: torch.Tensor, diffusion_time_embedding: Optional[torch.Tensor] = None):
        """Forward pass of the model.

        Args:
            x: input tensor
            diffusion_time_embedding: embedding of the diffusion time step

        Returns:
            Output tensor
        """
        h = self.input_block(x)
        h = self.middle_conv(h)

        if diffusion_time_embedding is not None:
            h = h + self.diffusion_step_projection(diffusion_time_embedding)

        h = self.output_block(h)

        if x.shape != h.shape:  # matching number of channels
            x = self.residual_projection(x)
        return (x + h) / math.sqrt(2.0)



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import warnings
from collections import Counter
from enum import Enum
from pathlib import Path
from typing import Dict, List, NewType, Optional, Union

from nemo.collections.common.tokenizers.tokenizer_spec import TokenizerSpec

__all__ = ['CharTokenizer']


NUMBER_OF_CHARACTERS_READ_BUFFER_SIZE = 10**7


class SpecialTokenString(Enum):
    MASK = 'mask'
    BOS = 'bos'
    EOS = 'eos'
    PAD = 'pad'
    SEP = 'sep'
    CLS = 'cls'
    UNK = 'unk'

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


SpecialTokenStringType = NewType('SpecialTokenString', SpecialTokenString)


class CharTokenizer(TokenizerSpec):
    rf"""
    Each character is a token.
    Args:
        vocab_file: path to file with vocabulary for a tokenizer. The file consists of valid Python string literals 
            separated by the new line character. Such literals must contain 1 character. Examples of valid Python 
            literals: ``'a'``, ``'\n'``, ``"'"``, ``'ж'``, ``'\u8976'``. Optionally the first line in the file can be a
            JSON dictionary of special tokens. The keys of the special tokens dictionary are ``'mask_token'``,
            ``'bos_token'`` and so on. Some special tokens names can be omitted in the special tokens dictionary line.
            A file ``vocab_file`` has to be in ``'utf-8'`` encoding.
        mask_token: mask token. The following is applicable to all special tokens. Parameter ``mask_token`` is used
            for adding mask token to vocabulary or for modification of mask token present in special tokens dictionary
            in the first line of file ``vocab_file``. Parameter ``mask_token`` can be either of type ``bool`` or a 
            ``str`` of length 1. 
            
            If ``mask_token`` is ``bool`` it has to be ``False``. If ``mask_token`` is ``True`` an exception is raised.
            If ``mask_token`` is ``False`` and ``mask_token`` is present in special tokens dictionary in vocabulary
            file ``vocab_file``, then ``mask_token`` is remove from special tokens dictionary.
            
            If the parameter ``mask_token`` is a string, then such strings in the input sequence are interpreted as
            mask tokens.
        bos_token: the beginning of sequence token. See more in ``mask_token`` parameter description.
        eos_token: the end of sequence token. Usually equal to sep_token. See more in ``mask_token`` parameter 
            description.
        pad_token: token to use for padding. See more in ``mask_token`` parameter description.
        sep_token: token used for separating sequences. See more in ``mask_token`` parameter description.
        cls_token: class token. Usually equal to bos_token. See more in ``mask_token`` parameter description.
        unk_token: token to use for unknown tokens. If the parameter ``unk_token`` is set and there is a character
            in the input of ``text_to_ids`` of ``text_to_tokens`` methods which is not in the vocabulary, then
            such an unknown character is tokenized into ``unk_token``. If the parameter ``unk_token`` is ``False``,
            then unknown tokens are discarded. See more in ``mask_token`` parameter description.
        special_token_to_prepend: special token to prepend to the output of ``text_to_ids`` of ``text_to_tokens``
            methods. This option can be used if you decide to add EOS and BOS tokens to the input on the stage of
            tokenization. Possible options are: {[None] + [e.value for e in SpecialTokenString]}.
        special_token_to_append: special token to append to the output of ``text_to_ids`` of ``text_to_tokens``
            methods. See more in the description of ``special_token_to_prepend`` parameter.
        special_tokens_to_remove_while_decoding: which special tokens are remove before detokenization. If this
            parameter equals ``'all'``, then all special tokens are removed. The parameter
            ``special_tokens_to_remove_while_decoding`` can also be a list of values from this set
            {set(e.value for e in SpecialTokenString)}.
    """

    def __init__(
        self,
        vocab_file: str,
        mask_token: Optional[Union[str, bool]] = None,
        bos_token: Optional[Union[str, bool]] = None,
        eos_token: Optional[Union[str, bool]] = None,
        pad_token: Optional[Union[str, bool]] = None,
        sep_token: Optional[Union[str, bool]] = None,
        cls_token: Optional[Union[str, bool]] = None,
        unk_token: Optional[Union[str, bool]] = None,
        special_token_to_prepend: Optional[SpecialTokenStringType] = None,
        special_token_to_append: Optional[SpecialTokenStringType] = None,
        special_tokens_to_remove_while_decoding: Union[List[SpecialTokenStringType], str] = 'all',
    ):
        vocab_file = Path(vocab_file).expanduser()
        with vocab_file.open(encoding='utf-8') as f:
            first_line = f.readline()
            if first_line[0] == '{':
                special_tokens_dict = json.loads(first_line)
                self.check_special_tokens_dict_from_file(special_tokens_dict, vocab_file)
                vocab_list = f.readlines()
            else:
                special_tokens_dict = {}
                vocab_list = [first_line] + f.readlines()
        special_tokens_dict = self.update_special_tokens_dict(
            special_tokens_dict, mask_token, bos_token, eos_token, pad_token, sep_token, cls_token, unk_token
        )
        for e in SpecialTokenString:
            name = e.value + '_token'
            setattr(self, name, special_tokens_dict[name] if name in special_tokens_dict else None)
        for k, v in special_tokens_dict.items():
            setattr(self, k, v)
        for value, name in [
            (special_token_to_prepend, 'special_token_to_prepend'),
            (special_token_to_append, 'special_token_to_append'),
        ]:
            self.check_special_token_name(name, value, special_tokens_dict)
            setattr(self, name, value + '_token' if isinstance(value, str) else value)
        self.vocab = {}
        count = 0
        for v in special_tokens_dict.values():
            self.vocab[v] = count
            count += 1
        for i, token in enumerate(vocab_list):

            if token[1:-1].startswith("\\"):
                token = token[1:-1].encode('ascii').decode('unicode_escape')[0]
            else:
                token = token[1:-1][0]

            self.check_token_from_file(token, vocab_file, i)
            if token not in self.vocab:
                self.vocab[token] = count
                count += 1
        self.inv_vocab = {v: k for k, v in self.vocab.items()}
        self.vocab_size = len(self.vocab)
        self.check_special_tokens_to_remove_while_decoding(
            special_tokens_to_remove_while_decoding, special_tokens_dict
        )
        self.special_token_ids_to_remove_while_decoding = (
            self.tokens_to_ids([v for v in special_tokens_dict.values()])
            if special_tokens_to_remove_while_decoding == 'all'
            else [getattr(self, e + '_id') for e in special_tokens_to_remove_while_decoding]
        )

    @classmethod
    def check_special_tokens_dict_from_file(cls, special_tokens_dict, vocab_file):
        for k, v in special_tokens_dict.items():
            if k[-6:] != '_token' or not SpecialTokenString.has_value(k[:-6]):
                raise ValueError(
                    f"Unsupported key {repr(k)} in special tokens dictionary in vocabulary file {vocab_file} "
                    f"(first line). Supported keys are {[e.value + '_token' for e in SpecialTokenString]}."
                )
            if not isinstance(v, str):
                raise ValueError(
                    f"Values of special tokens dictionary in vocabulary file {vocab_file} (first line) has to belong "
                    f"to type `str`, whereas type of item '{k}' value {repr(v)} is `{type(v)}`."
                )
            elif len(v) == 0:
                raise ValueError(
                    f"Values of special tokens dictionary in vocabulary file {vocab_file} (first line) has to not "
                    f"empty strings, whereas value of item '{k}' is an empty string."
                )
        cls.check_special_tokens_dict_for_duplicate_values(
            special_tokens_dict, f"Loaded from vocabulary file {vocab_file}"
        )

    @staticmethod
    def check_special_tokens_dict_for_duplicate_values(special_tokens_dict, err_msg_prefix):
        if len(special_tokens_dict) != len(set(special_tokens_dict.values())):
            tokens_with_equal_values = []
            duplicate_values = []
            for k, v in list(reversed(list(special_tokens_dict.items())))[:-1]:
                tokens = [k]
                for kk, vv in special_tokens_dict.items():
                    if kk == k:
                        break
                    if v == vv:
                        tokens.append(kk)
                if len(tokens) > 1:
                    duplicate_values.append(v)
                    tokens_with_equal_values.append(tokens)
            if duplicate_values:
                dup_values_msg = '. '.join(
                    [f"Tokens {t} have value '{v}'" for t, v in zip(tokens_with_equal_values, duplicate_values)]
                )
                raise ValueError(
                    err_msg_prefix + f" special tokens dictionary has duplicate values. " + dup_values_msg
                )

    @classmethod
    def update_special_tokens_dict(
        cls,
        init_special_tokens_dict: Dict[str, str],
        mask_token: Optional[Union[str, bool]] = None,
        bos_token: Optional[Union[str, bool]] = None,
        eos_token: Optional[Union[str, bool]] = None,
        pad_token: Optional[Union[str, bool]] = None,
        sep_token: Optional[Union[str, bool]] = None,
        cls_token: Optional[Union[str, bool]] = None,
        unk_token: Optional[Union[str, bool]] = None,
    ):
        special_tokens_dict = init_special_tokens_dict.copy()
        for value, name in zip(
            [pad_token, unk_token, bos_token, eos_token, sep_token, mask_token, cls_token],
            ['pad_token', 'unk_token', 'bos_token', 'eos_token', 'sep_token', 'mask_token', 'cls_token'],
        ):
            if value is not None:
                if isinstance(value, bool):
                    if value:
                        raise ValueError(
                            f"If `CharTokenizer` constructor parameter `{name}` is `bool` it has to be `False`"
                        )
                    else:
                        if name in special_tokens_dict:
                            del special_tokens_dict[name]
                        else:
                            warnings.warn(
                                f"Cannot remove special token `{name}` since it is not in special tokens dictionary "
                                f"{special_tokens_dict}."
                            )
                elif not isinstance(value, str):
                    raise ValueError(
                        f"`CharTokenizer` constructor parameter `{name}` has to be either `False` or belong to type "
                        f"`str`, whereas type of `{name}` is `{type(value)}`."
                    )
                else:
                    special_tokens_dict[name] = value
        cls.check_special_tokens_dict_for_duplicate_values(
            special_tokens_dict,
            "After updating special tokens dictionary with tokens passed in `CharTokenizer` constructor parameters",
        )
        return special_tokens_dict

    @staticmethod
    def check_token_from_file(token, vocab_file, line_i):
        if not isinstance(token, str) or isinstance(token, str) and len(token) != 1:
            raise ValueError(
                f"Each line in vocabulary have to be a Python string literal containing 1 character. "
                f"Encountered {repr(token)} on line {line_i} in file {vocab_file}."
            )

    @staticmethod
    def check_special_token_name(parameter_name, value, special_tokens_dict):
        if value is not None:
            if not SpecialTokenString.has_value(value):
                raise ValueError(
                    f"Value {repr(value)} of parameter `{parameter_name}` is wrong. Supported values are "
                    f"{[e.value for e in SpecialTokenString]}."
                )
            elif value + '_token' not in special_tokens_dict:
                raise ValueError(
                    f"You should provide `{value + '_token'}` parameter to `CharTokenizer` constructor if "
                    f"you wish to pass token {repr(value)} in parameter `{parameter_name}`."
                )

    @staticmethod
    def check_special_tokens_to_remove_while_decoding(special_tokens_to_remove_while_decoding, special_tokens_dict):
        if isinstance(special_tokens_to_remove_while_decoding, list):
            for i, value in enumerate(special_tokens_to_remove_while_decoding):
                if not SpecialTokenString.has_value(value):
                    raise ValueError(
                        f'Wrong element with value {repr(value)} in position {i} of parameter '
                        f'`special_tokens_to_remove_while_decoding` of `CharTokenizer` constructor. Supported values '
                        f'are {[e.value for e in SpecialTokenString]}.'
                    )
                elif value + '_token' not in special_tokens_dict:
                    raise ValueError(
                        f"You should provide `{value + '_token'}` parameter to `CharTokenizer` constructor if "
                        f"you wish to pass token {repr(value)} in parameter `special_tokens_to_remove_while_decoding`. "
                        f"`{value + '_token'}` was detected in position {i} in "
                        f"`special_tokens_to_remove_while_decoding`."
                    )
        elif (
            isinstance(special_tokens_to_remove_while_decoding, str)
            and special_tokens_to_remove_while_decoding != 'all'
            or not isinstance(special_tokens_to_remove_while_decoding, str)
        ):
            raise ValueError(
                f"Parameter `special_tokens_to_remove_while_decoding` of `CharTokenizer` constructor has to be "
                f"equal to a string 'all' or be a list of values from set {set(e.value for e in SpecialTokenString)} "
                f"whereas `special_tokens_to_remove_while_decoding={repr(special_tokens_to_remove_while_decoding)}`"
            )

    def text_to_tokens(self, text: str) -> List[str]:
        token_candidates = [char for char in text]
        tokens = []
        if self.special_token_to_prepend is not None:
            tokens.append(getattr(self, self.special_token_to_prepend))
        for i, token in enumerate(token_candidates):
            if token in self.vocab:
                tokens.append(token)
            elif self.unk_token is not None:
                tokens.append(self.unk_token)
            else:
                warnings.warn(
                    f"Character {repr(token)} in position {i} is not present in vocabulary and no `<UNK>` token was "
                    f"set. Character {repr(token)} is discarded."
                )
        if self.special_token_to_append is not None:
            tokens.append(getattr(self, self.special_token_to_append))
        return tokens

    def tokens_to_text(self, tokens: List[str]) -> str:
        return self.ids_to_text(self.tokens_to_ids(tokens))

    def text_to_ids(self, text: str) -> List[int]:
        ids = [self.vocab[token] for token in self.text_to_tokens(text)]
        return ids

    def ids_to_text(self, ids: List[int]) -> str:
        ids_ = [id_ for id_ in ids if id_ not in self.special_token_ids_to_remove_while_decoding]
        return "".join(self.ids_to_tokens(ids_))

    def tokens_to_ids(self, tokens: List[str]) -> List[int]:
        return [self.vocab[token] for token in tokens]

    def token_to_id(self, token: str) -> int:
        return self.vocab[token]

    def ids_to_tokens(self, ids: List[int]) -> List[str]:
        return [self.inv_vocab[id] for id in ids]

    @staticmethod
    def check_special_token_id_getting(special_token, id_name):
        if special_token is None:
            token_param = id_name[:-3] + '_token'
            raise ValueError(
                f"Cannot return `{id_name}` since `{token_param}` is not set. To obtain `{id_name}` you need to pass "
                f"parameter `{token_param}` to `CharTokenizer` constructor."
            )

    @property
    def pad_id(self):
        self.check_special_token_id_getting(self.pad_token, 'pad_id')
        return self.vocab[self.pad_token]

    @property
    def bos_id(self):
        self.check_special_token_id_getting(self.bos_token, 'bos_id')
        return self.vocab[self.bos_token]

    @property
    def eos_id(self):
        self.check_special_token_id_getting(self.eos_token, 'eos_id')
        return self.vocab[self.eos_token]

    @property
    def unk_id(self):
        self.check_special_token_id_getting(self.unk_token, 'unk_id')
        return self.vocab[self.unk_token]

    @property
    def mask_id(self):
        self.check_special_token_id_getting(self.mask_token, 'mask_id')
        return self.vocab[self.mask_token]

    @property
    def sep_id(self):
        self.check_special_token_id_getting(self.sep_token, 'sep_id')
        return self.vocab[self.sep_token]

    @property
    def cls_id(self):
        self.check_special_token_id_getting(self.cls_token, 'cls_id')
        return self.vocab[self.cls_token]

    @staticmethod
    def create_special_tokens_dict(
        mask_token: Optional[str] = None,
        bos_token: Optional[str] = None,
        eos_token: Optional[str] = None,
        pad_token: Optional[str] = None,
        sep_token: Optional[str] = None,
        cls_token: Optional[str] = None,
        unk_token: Optional[str] = None,
    ):
        special_tokens_dict = {}
        for value, name in zip(
            [pad_token, unk_token, bos_token, eos_token, sep_token, mask_token, cls_token],
            ['pad_token', 'unk_token', 'bos_token', 'eos_token', 'sep_token', 'mask_token', 'cls_token'],
        ):
            if value is not None:
                if not isinstance(value, str):
                    raise ValueError(
                        f"The type of parameter `{name}` has to be `None` or `str`, found `{type(value)}`"
                    )
                elif len(value) == 0:
                    raise ValueError(f"If the parameter `{name}` is `str`, then its length has to be nonzero.")
                elif value in special_tokens_dict.values():
                    other_name = None
                    for k, v in special_tokens_dict.items():
                        if v == value:
                            other_name = k
                    raise ValueError(
                        f"The value {repr(value)} of special token `{name}` is the same as the value of special token "
                        f"`{other_name}`."
                    )
                special_tokens_dict[name] = value
        return special_tokens_dict

    @staticmethod
    def check_characters_to_exclude_from_vocabulary(characters_to_exclude_from_vocabulary):
        for i, char in enumerate(characters_to_exclude_from_vocabulary):
            if not isinstance(char, str):
                raise ValueError(
                    f"Character to exclude from vocabulary has to `str`, whereas an element in position {i} is of "
                    f"type `{type(char)}`."
                )
            elif len(char) != 1:
                raise ValueError(
                    f"A length of an element of `characters_to_exclude_from_vocabulary` parameter has to be 1. "
                    f"The length of an element in position {i} is {len(char)}."
                )

    @staticmethod
    def check_text_and_text_file_name(text, text_file_name):
        if text is None and text_file_name is None:
            raise ValueError(
                f'Exactly one of parameters `text` and `text_file_name` should be provided whereas both parameters '
                f'are `None`.'
            )
        if text is not None and text_file_name is not None:
            raise ValueError(
                f"Exactly one of parameters `text` and `text_file_name` has to be provided, whereas both parameters "
                f"are not `None`."
            )
        if text is not None:
            if not isinstance(text, str):
                raise ValueError(
                    f"Parameter `text` has to be of type `str`, whereas it belongs to type `{type(text)}`."
                )

    @classmethod
    def build_vocab(
        cls,
        save_path: Union[str, bytes, os.PathLike],
        text: Optional[str] = None,
        text_file_name: Optional[Union[str, bytes, os.PathLike]] = None,
        characters_to_exclude: Optional[List[str]] = None,
        vocab_size: int = None,
        mask_token: Optional[str] = None,
        bos_token: Optional[str] = None,
        eos_token: Optional[str] = None,
        pad_token: Optional[str] = None,
        sep_token: Optional[str] = None,
        cls_token: Optional[str] = None,
        unk_token: Optional[str] = None,
    ):
        """
        Creates character vocabulary and saves it to file ``save_path``. You should provide one of parameters ``text``
        and ``text_file_name``. The format of created character vocabulary file is following:
        ```
        {['mask_token': "ANY NON EMPTY STRING", ]['bos_token': "ANY NON EMPTY STRING", ] and so on}
        ' '
        'e'
        ...
        ```
        The first line is a JSON which contains special tokens. This special token are set using parameters
        ``mas_token``, ``bos_token``, ``eos_token``, ``pad_token``, ``sep_token``, ``cls_token``, ``unk_token``.
        Other lines in created vocabulary file are Python string literals containing one character each.

        Args:
            save_path: path to the output text file. If ``save_path`` parent directory does not exist it will be created
            text: string which characters are used for vocabulary creation.
            text_file_name: path to a file which characters are used for vocabulary creation. Use this parameter if
                the text in file is too large to be loaded in memory.
            characters_to_exclude: a list of characters which will not be added to vocabulary.
            vocab_size: vocabulary size. If this parameter is set only most frequent ``vocab_size`` characters are added
                to vocabulary.
            mask_token: mask token
            bos_token: the beginning of sequence token
            eos_token: the end of sequence token. Usually equal to sep_token.
            pad_token: token to use for padding.
            sep_token: token used for separating sequences.
            cls_token: class token. Usually equal to bos_token.
            unk_token: token to use for unknown tokens. If the parameter ``unk_token`` is set and there is a character
                in the input of ``text_to_ids`` of ``text_to_tokens`` methods which is not in the vocabulary, then
                such an unknown character is tokenized into ``unk_token``. If the parameter ``unk_token`` is ``False``,
                then unknown tokens are discarded.
        """
        special_tokens_dict = cls.create_special_tokens_dict(
            mask_token, bos_token, eos_token, pad_token, sep_token, cls_token, unk_token
        )
        if characters_to_exclude is None:
            characters_to_exclude = []
        else:
            cls.check_characters_to_exclude_from_vocabulary(characters_to_exclude)
        cls.check_text_and_text_file_name(text, text_file_name)
        if text is not None:
            counter = Counter(text)
        else:
            assert text_file_name is not None
            text_file_name = Path(text_file_name).expanduser()
            counter = Counter()
            with text_file_name.open(encoding='utf-8') as f:
                while True:
                    segment = f.read(NUMBER_OF_CHARACTERS_READ_BUFFER_SIZE)
                    if not segment:
                        break
                    counter.update(segment)
        for char in characters_to_exclude:
            if char in counter:
                del counter[char]
        save_path = Path(save_path).expanduser()
        save_path.parent.mkdir(exist_ok=True, parents=True)
        with save_path.open('w', encoding='utf-8') as f:
            f.write(json.dumps(special_tokens_dict) + '\n')
            if vocab_size is None:
                for c, _ in sorted(counter.items(), key=lambda x: -x[1]):
                    f.write(repr(c) + '\n')
            else:
                vocab_size -= len(special_tokens_dict)
                for i, (c, _) in enumerate(sorted(counter.items(), key=lambda x: -x[1])):
                    if i < vocab_size:
                        f.write(repr(c) + '\n')
                    else:
                        break



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile

import pytest
import torch
from omegaconf import DictConfig

from nemo.collections.tts.models import WaveGlowModel
from nemo.core.classes import typecheck

mcfg = DictConfig(
    {
        "_target_": "nemo.collections.tts.modules.waveglow.WaveGlowModule",
        "n_flows": 12,
        "n_group": 8,
        "n_mel_channels": 80,
        "n_early_every": 4,
        "n_early_size": 2,
        "n_wn_channels": 512,
        "n_wn_layers": 8,
        "wn_kernel_size": 3,
    }
)

pcfg = DictConfig(
    {
        "_target_": "nemo.collections.asr.parts.preprocessing.features.FilterbankFeatures",
        "dither": 0.0,
        "nfilt": 80,
        "stft_conv": False,
    }
)

wcfg = DictConfig({"waveglow": mcfg, "sigma": 1.0, "preprocessor": pcfg,})


def input_example(sz):
    mel = torch.randn(1, 1, 80, sz).cuda().half()
    z = torch.randn(1, 8, sz * 256 // 8, 1).cuda().half()
    return (
        mel,
        z,
    )


def taco2wg(spec, z):
    spec = spec.permute(0, 3, 2, 1).contiguous()
    return spec.view(spec.size(0), spec.size(1), -1), z.view(z.size(0), z.size(1), -1)


# Wrapper method to convert Jasper's Taco2 output to WG input and call inference
def forward_wrapper(self, spec, z=None):
    spec, z = taco2wg(spec, z)
    audio = self.waveglow.norm_dist_to_audio(spec=spec, sigma=1.0, z=z)
    return audio


class TestWaveGlow:
    @pytest.mark.pleasefixme
    @pytest.mark.run_only_on('GPU')
    @pytest.mark.unit
    def test_export_to_onnx(self):
        model = WaveGlowModel(wcfg)
        model = model.cuda().half()
        typecheck.set_typecheck_enabled(enabled=False)
        with tempfile.TemporaryDirectory() as tmpdir, model.nemo_infer():
            tmp_file_name = os.path.join(tmpdir, "waveglow.onnx")

            n_mels = 80
            # Test export.
            inp = input_example(n_mels)
            inp1 = taco2wg(*inp)
            inp2 = inp1
            res1 = model.waveglow(*inp1)
            res2 = model.waveglow(*inp2)
            assert torch.allclose(res1, res2, rtol=0.01, atol=0.1)
            WaveGlowModel.forward_for_export = forward_wrapper
            model.export(
                tmp_file_name, input_example=inp, verbose=False, check_trace=False, do_constant_folding=True,
            )


if __name__ == "__main__":
    t = TestWaveGlow()
    t.test_export_to_onnx()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from typing import Callable, Optional, Union

import lightning.pytorch as pl
import nemo_run as run
import torch
from lightning.pytorch.callbacks.callback import Callback
from megatron.core.distributed import DistributedDataParallelConfig
from megatron.core.transformer.transformer_config import TransformerConfig

from nemo import lightning as nl
from nemo.collections import vlm
from nemo.collections.llm.api import pretrain
from nemo.collections.llm.gpt.model.llama import Llama3Config8B
from nemo.collections.llm.recipes.log.default import default_log, default_resume, tensorboard_logger
from nemo.collections.llm.recipes.optim.adam import distributed_fused_adam_with_cosine_annealing
from nemo.collections.llm.recipes.precision.mixed_precision import bf16_mixed
from nemo.collections.vlm.neva.data.mock import MockDataModule
from nemo.lightning.pytorch.callbacks.garbage_collection import GarbageCollectionCallback
from nemo.lightning.pytorch.callbacks.megatron_comm_overlap import MegatronCommOverlapCallback
from nemo.utils.exp_manager import TimingCallback

NAME = "neva_llama3_8b"


@dataclass
class NevaConfig8B(vlm.NevaConfig):
    """NeVA (CLIP-ViT-L + LLaMa38B) Config"""

    from transformers import PretrainedConfig

    language_transformer_config: TransformerConfig = field(default_factory=lambda: Llama3Config8B(seq_length=8192))
    vision_transformer_config: Union[TransformerConfig, PretrainedConfig] = field(
        default_factory=lambda: vlm.HFCLIPVisionConfig(
            pretrained_model_name_or_path="openai/clip-vit-large-patch14-336"
        )
    )
    vision_projection_config: TransformerConfig = field(
        default_factory=lambda: vlm.MultimodalProjectorConfig(input_size=1024, hidden_size=4096, ffn_hidden_size=4096)
    )

    freeze_language_model: bool = False
    freeze_vision_model: bool = True
    freeze_vision_projection: bool = False


@run.cli.factory(name=NAME)
def model(
    seq_length: Optional[int] = 8192,
) -> run.Config[pl.LightningModule]:
    """
    Factory function to create a NeVA (CLIP-ViT-L + LLaMa3 8B) model configuration.

    Returns:
        run.Config[pl.LightningModule]: Configuration for the NeVA (CLIP-ViT-L + LLaMa3 8B) model.
    """
    model_config = run.Config(vlm.NevaModel, config=run.Config(NevaConfig8B))
    return model_config


def trainer(
    tensor_parallelism: int = 1,
    pipeline_parallelism: int = 1,
    pipeline_parallelism_type: Optional[torch.dtype] = None,
    virtual_pipeline_parallelism: Optional[int] = None,
    context_parallelism: int = 2,
    sequence_parallelism: bool = False,
    num_nodes: int = 1,
    num_gpus_per_node: int = 8,
    max_steps: int = 1168251,
    callbacks: Optional[list[run.Config[Callback]]] = None,
) -> run.Config[nl.Trainer]:
    """
    Configure the NeMo Lightning Trainer for NeVA (CLIP-ViT-L + LLaMa3 8B) model.

    This function sets up the distributed training strategy and other training parameters.

    Args:
        tensor_parallelism (int): Degree of tensor model parallelism.
        pipeline_parallelism (int): Degree of pipeline model parallelism.
        pipeline_parallelism_type (Optional[torch.dtype]): Data type for pipeline parallelism.
        virtual_pipeline_parallelism (Optional[int]): Size of virtual pipeline parallelism.
        context_parallelism (int): Degree of context parallelism.
        sequence_parallelism (bool): Whether to use sequence parallelism.
        num_nodes (int): Number of compute nodes to use.
        num_gpus_per_node (int): Number of GPUs per node.
        max_steps (int): Maximum number of training steps.
        callbacks (Optional[list[run.Config[Callback]]]): List of callback configurations.

    Returns:
        run.Config[nl.Trainer]: Configuration for the NeMo Lightning Trainer.

    Examples:
        CLI usage:
            $ nemo llm finetune trainer=neva_llama3_8b ...

        Python API usage:
            >>> trainer_config = trainer(num_nodes=2, num_gpus_per_node=8)
            >>> print(trainer_config)

    Note:
        For more information on distributed training strategies, refer to the
        NeMo documentation on multi-GPU and multi-node training.
    """
    strategy = run.Config(
        nl.MegatronStrategy,
        tensor_model_parallel_size=tensor_parallelism,
        pipeline_model_parallel_size=pipeline_parallelism,
        pipeline_dtype=pipeline_parallelism_type,
        virtual_pipeline_model_parallel_size=virtual_pipeline_parallelism,
        context_parallel_size=context_parallelism,
        sequence_parallel=sequence_parallelism,
        gradient_as_bucket_view=True,
        ckpt_async_save=True,
        ckpt_parallel_load=True,
        ddp=run.Config(
            DistributedDataParallelConfig,
            check_for_nan_in_grad=True,
            grad_reduce_in_fp32=False,
            overlap_grad_reduce=False,
            overlap_param_gather=True,
            average_in_collective=True,
        ),
    )

    trainer = run.Config(
        nl.Trainer,
        accelerator="gpu",
        accumulate_grad_batches=1,
        callbacks=callbacks,
        devices=num_gpus_per_node,
        limit_test_batches=50,
        limit_val_batches=32,
        log_every_n_steps=10,
        max_steps=max_steps,
        num_nodes=num_nodes,
        plugins=bf16_mixed(),
        strategy=strategy,
        use_distributed_sampler=False,
        val_check_interval=2000,
    )

    return trainer


@run.cli.factory(target=pretrain, name=NAME)
def finetune_recipe(
    dir: Optional[str] = None,
    name: str = "default",
    num_nodes: int = 1,
    num_gpus_per_node: int = 8,
    seq_length: int = 8192,
    performance_mode: bool = False,
    packed_sequence: bool = False,
    fn: Callable = pretrain,
) -> run.Partial:
    """
    Create a finetuning recipe for NeVA (CLIP-ViT-L + LLaMa3 8B) model.
    Using pre-train fn targets to use appropriate llm api call.

    This function sets up a complete configuration for finetuning, including
    model, trainer, data, logging, optimization, and resumption settings.

    Args:
        dir (Optional[str]): Directory for saving logs and checkpoints.
        name (str): Name of the finetuning run.
        num_nodes (int): Number of compute nodes to use.
        num_gpus_per_node (int): Number of GPUs per node.
        performance_mode (bool): If true, enables optimizations for maximum performance.
        fn (Callable): The finetuning function to use.

    Returns:
        run.Partial: Partial configuration for fine-tuning.

    Examples:
        CLI usage:
            $ nemo llm finetune --factory neva_llama3_8b

        Python API usage:
            >>> recipe = finetune_recipe(name="neva_llama3_8b_finetune", num_nodes=2)
            >>> print(recipe)

    """
    recipe = run.Partial(
        fn,
        model=model(seq_length=seq_length),
        trainer=trainer(
            num_nodes=num_nodes,
            num_gpus_per_node=num_gpus_per_node,
            callbacks=[run.Config(TimingCallback)],
        ),
        data=run.Config(
            MockDataModule,
            seq_length=seq_length,
            global_batch_size=512,
            micro_batch_size=1,
            num_workers=4,
            packed_sequence=packed_sequence,
        ),
        log=default_log(dir=dir, name=name, tensorboard_logger=tensorboard_logger(name=name)),
        optim=distributed_fused_adam_with_cosine_annealing(max_lr=3e-4),
        resume=default_resume(),
    )

    if performance_mode:
        recipe = finetune_performance_optimizations(recipe)

    return recipe


def finetune_performance_optimizations(recipe: run.Partial) -> run.Partial:
    """
    Create a performance-optimized finetuning recipe for Llama3 8B model.

    This method enables performance optimizations that may not be suitable for all use cases.
    It builds upon the standard finetuning recipe and adds additional performance enhancements.

    Args:
        recipe (run.Partial): Base finetune recipe to which performance optimizations will be added

    Returns:
        run.Partial: Partial configuration for performance-optimized finetuning.

    Note:
        Use this method with caution and only when you need maximum performance.
        It may not be suitable for all hardware configurations or use cases.
    """
    if not hasattr(recipe.trainer, "callbacks") or recipe.trainer.callbacks is None:
        recipe.trainer.callbacks = []

    if recipe.trainer.strategy.tensor_model_parallel_size > 1:
        recipe.trainer.callbacks.append(
            run.Config(
                MegatronCommOverlapCallback,
                tp_comm_overlap=True,
            )
        )
    else:
        recipe.trainer.callbacks.append(
            run.Config(
                MegatronCommOverlapCallback,
                tp_comm_overlap=False,
            )
        )

    recipe.trainer.callbacks.append(run.Config(TimingCallback))
    recipe.trainer.callbacks.append(
        run.Config(
            GarbageCollectionCallback,
            100,
            100,
        )
    )

    recipe.trainer.plugins.grad_reduce_in_fp32 = False

    return recipe



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from torch import nn

from nemo.collections.multimodal.speech_cv.parts.submodules.permute import Permute
from nemo.core.classes import NeuralModule, typecheck

try:
    import torchvision

    TORCHVISION_AVAILABLE = True
except (ImportError, ModuleNotFoundError):
    TORCHVISION_AVAILABLE = False


class VideoPreprocessor(NeuralModule):

    """ Video Pre-processing

    args:
        grayscale: convert images to grayscale
        normalize: normalize videos
        resize: resize videos
        resize_size: output image size for resize
        norm_mean: normalize mean
        norm_std: normalize std
    
    """

    def __init__(self, grayscale, normalize, resize, resize_size, norm_mean, norm_std):
        super().__init__()

        # Params
        self.grayscale = grayscale
        self.normalize = normalize
        self.resize = resize
        self.resize_size = resize_size
        self.norm_mean = norm_mean
        self.norm_std = norm_std

        self.transforms = nn.ModuleList()

        # Convert float32 [0:255] -> [0:1]
        if TORCHVISION_AVAILABLE:
            self.transforms.append(torchvision.transforms.ConvertImageDtype(dtype=torch.float32))
        else:
            raise Exception("ConvertImageDtype transform requires torchvision")

        # Convert Channels First
        self.transforms.append(Permute(dims=(0, 4, 1, 2, 3)))  # (B, T, H, W, C) -> (B, C, T, H, W)

        # Resize
        if self.resize:
            self.transforms.append(ResizeVideo(self.resize_size))  # (B, C, T, H, W) -> (B, C, T, H', W')

        # Grayscale
        if self.grayscale:
            if TORCHVISION_AVAILABLE:
                self.transforms.append(
                    nn.Sequential(
                        Permute(dims=(0, 2, 1, 3, 4)),  # (B, C, T, H, W) -> (B, T, C, H, W)
                        torchvision.transforms.Grayscale(),
                        Permute(dims=(0, 2, 1, 3, 4)),  # (B, T, C, H, W) -> (B, C, T, H, W)
                    )
                )
            else:
                raise Exception("Grayscale transform requires torchvision")

        # Normalize
        if self.normalize:
            self.transforms.append(NormalizeVideo(mean=norm_mean, std=norm_std))

    @typecheck()
    @torch.no_grad()
    def forward(self, input_signal, length):

        for transform in self.transforms:
            input_signal = transform(input_signal)

        return input_signal, length


class NormalizeVideo(NeuralModule):
    def __init__(self, mean, std):
        super().__init__()

        self.register_buffer(
            "mean", torch.tensor(mean, dtype=torch.float32).reshape(len(mean), 1, 1, 1), persistent=False
        )
        self.register_buffer(
            "std", torch.tensor(std, dtype=torch.float32).reshape(len(std), 1, 1, 1), persistent=False
        )

    def forward(self, x):

        x = (x - self.mean) / self.std

        return x


class ResizeVideo(NeuralModule):
    def __init__(self, size):
        super().__init__()

        self.size = size
        if TORCHVISION_AVAILABLE:
            self.resize = torchvision.transforms.Resize(size=self.size)
        else:
            raise Exception("Resize transform requires torchvision")

    def forward(self, x):

        # (B, C, T, H, W)
        if x.dim() == 5:

            B, C = x.shape[:2]
            x = x.flatten(start_dim=0, end_dim=1)
            x = self.resize(x)
            x = x.reshape((B, C) + x.shape[1:])

        # (C, T, H, W)
        elif x.dim() == 4:
            x = self.resize(x)

        return x



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from typing import Any, Dict, List, Optional

import torch
from omegaconf import DictConfig, open_dict
from pytorch_lightning import Trainer

from nemo.collections.asr.data.audio_to_text_lhotse_speaker import LhotseSpeechToTextSpkBpeDataset
from nemo.collections.asr.models.rnnt_bpe_models import EncDecRNNTBPEModel
from nemo.collections.asr.parts.mixins import TranscribeConfig
from nemo.collections.asr.parts.mixins.multitalker_asr_mixins import SpeakerKernelMixin
from nemo.collections.common.data.lhotse import get_lhotse_dataloader_from_config
from nemo.core.classes.common import PretrainedModelInfo


class EncDecMultiTalkerRNNTBPEModel(EncDecRNNTBPEModel, SpeakerKernelMixin):
    """Base class for encoder decoder RNNT-based models with subword tokenization."""

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        super().__init__(cfg=cfg, trainer=trainer)
        # Initialize speaker kernel functionality from mixin
        self._init_speaker_kernel_config(cfg)

    @classmethod
    def list_available_models(cls) -> List[PretrainedModelInfo]:
        """
        This method returns a list of pre-trained model which can be instantiated directly from NVIDIA's NGC cloud.

        Returns:
            List of available pre-trained models.
        """
        results = []

        return results

    def _setup_dataloader_from_config(self, config: Optional[Dict]):
        if config.get("use_lhotse"):
            # Use open_dict to allow dynamic key addition
            with open_dict(config):
                config.global_rank = self.global_rank
                config.world_size = self.world_size

            return get_lhotse_dataloader_from_config(
                config,
                global_rank=self.global_rank,
                world_size=self.world_size,
                dataset=LhotseSpeechToTextSpkBpeDataset(
                    cfg=config,
                    tokenizer=self.tokenizer,
                ),
            )

    def training_step(self, batch, batch_nb):
        """Training step with speaker targets."""
        signal, signal_len, transcript, transcript_len, *additional_args = batch
        spk_targets, bg_spk_targets = additional_args

        self.set_speaker_targets(spk_targets, bg_spk_targets)

        batch = (signal, signal_len, transcript, transcript_len)

        return super().training_step(batch, batch_nb)

    def validation_pass(self, batch, batch_idx, dataloader_idx=0):
        """Validation pass with speaker targets."""
        signal, signal_len, transcript, transcript_len, *additional_args = batch
        spk_targets, bg_spk_targets = additional_args

        self.set_speaker_targets(spk_targets, bg_spk_targets)

        batch = (signal, signal_len, transcript, transcript_len)

        return super().validation_pass(batch, batch_idx, dataloader_idx)

    def _transcribe_forward(self, batch: Any, trcfg: TranscribeConfig):
        """Transcribe forward with speaker targets."""
        signal, signal_len, transcript, transcript_len, *additional_args = batch
        spk_targets, bg_spk_targets = additional_args

        self.set_speaker_targets(spk_targets, bg_spk_targets)

        batch = (signal, signal_len, transcript, transcript_len)

        return super()._transcribe_forward(batch, trcfg)

    def _setup_transcribe_dataloader(self, config: Dict) -> 'torch.utils.data.DataLoader':
        """
        Setup function for a temporary data loader which wraps the provided audio file.

        Args:
            config: A python dictionary which contains the following keys:
            paths2audio_files: (a list) of paths to audio files. The files should be relatively short fragments. \
                Recommended length per file is between 5 and 25 seconds.
            batch_size: (int) batch size to use during inference. \
                Bigger will result in better throughput performance but would use more memory.
            temp_dir: (str) A temporary directory where the audio manifest is temporarily
                stored.

        Returns:
            A pytorch DataLoader for the given audio file(s).
        """
        if 'dataset_manifest' in config:
            manifest_filepath = config['dataset_manifest']
            batch_size = config['batch_size']
        else:
            manifest_filepath = os.path.join(config['temp_dir'], 'manifest.json')
            batch_size = min(config['batch_size'], len(config['paths2audio_files']))

        dl_config = {
            'manifest_filepath': manifest_filepath,
            'sample_rate': self.preprocessor._sample_rate,
            'batch_size': batch_size,
            'shuffle': False,
            'num_workers': config.get('num_workers', min(batch_size, os.cpu_count() - 1)),
            'pin_memory': True,
            'use_lhotse': config.get('use_lhotse', True),
            'use_bucketing': False,
            'channel_selector': config.get('channel_selector', None),
            'inference_mode': self.cfg.test_ds.get('inference_mode', True),
            'fixed_spk_id': config.get('fixed_spk_id', None),
        }

        if config.get("augmentor"):
            dl_config['augmentor'] = config.get("augmentor")

        temporary_datalayer = self._setup_dataloader_from_config(config=DictConfig(dl_config))

        return temporary_datalayer
